"""Convenience classes for configuring Vizier Study Configs and Search Spaces.

This module contains several classes, used to access/build Vizier StudyConfig
protos:
  * `StudyConfig` class is the main class, which:
  1) Allows to easily build Vizier StudyConfig protos via a convenient
     Python API.
  2) Can be initialized from an existing StudyConfig proto, to enable easy
     Pythonic accessors to information contained in StudyConfig protos,
     and easy field editing capabilities.

  * `SearchSpace` and `SearchSpaceSelector` classes deals with Vizier search
    spaces. Both flat spaces and conditional parameters are supported.
"""
import collections
import copy
import enum
from typing import Dict, List, Optional, Sequence, Tuple, Union

import attr
from vizier._src.pyvizier.oss import automated_stopping
from vizier._src.pyvizier.oss import metadata_util
from vizier._src.pyvizier.oss import proto_converters
from vizier._src.pyvizier.shared import base_study_config
from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import parameter_config
from vizier._src.pyvizier.shared import trial
from vizier.service import study_pb2

################### PyTypes ###################
ScaleType = parameter_config.ScaleType
ExternalType = parameter_config.ExternalType
# A sequence of possible internal parameter values.
MonotypeParameterSequence = parameter_config.MonotypeParameterSequence
# Possible types for trial parameter values after cast to external types.
ParameterValueSequence = Union[trial.ParameterValueTypes, Sequence[int],
                               Sequence[float], Sequence[str], Sequence[bool]]

################### Enums ###################

# Values should NEVER be removed from ObjectiveMetricGoal, only added.
ObjectiveMetricGoal = base_study_config.ObjectiveMetricGoal


class Algorithm(enum.Enum):
  """Valid Values for StudyConfig.Algorithm."""
  ALGORITHM_UNSPECIFIED = study_pb2.StudySpec.Algorithm.ALGORITHM_UNSPECIFIED
  GAUSSIAN_PROCESS_BANDIT = study_pb2.StudySpec.Algorithm.GAUSSIAN_PROCESS_BANDIT
  GRID_SEARCH = study_pb2.StudySpec.Algorithm.GRID_SEARCH
  RANDOM_SEARCH = study_pb2.StudySpec.Algorithm.RANDOM_SEARCH
  NSGA2 = study_pb2.StudySpec.Algorithm.NSGA2
  EMUKIT_GP_EI = study_pb2.StudySpec.Algorithm.EMUKIT_GP_EI


class ObservationNoise(enum.Enum):
  """Valid Values for StudyConfig.ObservationNoise."""
  OBSERVATION_NOISE_UNSPECIFIED = study_pb2.StudySpec.ObservationNoise.OBSERVATION_NOISE_UNSPECIFIED
  LOW = study_pb2.StudySpec.ObservationNoise.LOW
  HIGH = study_pb2.StudySpec.ObservationNoise.HIGH


SearchSpaceSelector = base_study_config.SearchSpaceSelector

################### Main Class ###################
#
# A StudyConfig object can be initialized:
# (1) From a StudyConfig proto using StudyConfig.from_proto():
#     study_config_proto = study_pb2.StudySpec(...)
#     study_config = pyvizier.StudyConfig.from_proto(study_config_proto)
#     # Attributes can be modified.
#     study_config.metadata['metadata_key'] = 'metadata_value'
#     new_proto = study_config.to_proto()
#
# (2) By directly calling __init__ and setting attributes:
#     study_config = pyvizier.StudyConfig(
#       metric_information=[pyvizier.MetricInformation(
#         name='accuracy', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)],
#       search_space=SearchSpace.from_proto(proto),
#     )
#     # OR:
#     study_config = pyvizier.StudyConfig()
#     study_config.metric_information.append(
#        pyvizier.MetricInformation(
#          name='accuracy', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE))
#
#     # Since building a search space is more involved, get a reference to the
#     # search space, and add parameters to it.
#     root = study_config.search_space.select_root()
#     root.add_float_param('learning_rate', 0.001, 1.0,
#       scale_type=pyvizier.ScaleType.LOG)
#


@attr.define(frozen=False, init=True, slots=True, kw_only=True)
class StudyConfig(base_study_config.ProblemStatement):
  """A builder and wrapper for study_pb2.StudySpec proto."""

  search_space: base_study_config.SearchSpace = attr.field(
      init=True,
      factory=base_study_config.SearchSpace,
      validator=attr.validators.instance_of(base_study_config.SearchSpace),
      on_setattr=attr.setters.validate)

  algorithm: Algorithm = attr.field(
      init=True,
      validator=attr.validators.instance_of(Algorithm),
      on_setattr=[attr.setters.convert, attr.setters.validate],
      default=Algorithm.GAUSSIAN_PROCESS_BANDIT,
      kw_only=True)

  # TODO: This name/type combo is confusing.
  metric_information: base_study_config.MetricsConfig = attr.field(
      init=True,
      factory=base_study_config.MetricsConfig,
      converter=base_study_config.MetricsConfig,
      validator=attr.validators.instance_of(base_study_config.MetricsConfig),
      kw_only=True)

  observation_noise: ObservationNoise = attr.field(
      init=True,
      validator=attr.validators.instance_of(ObservationNoise),
      on_setattr=attr.setters.validate,
      default=ObservationNoise.OBSERVATION_NOISE_UNSPECIFIED,
      kw_only=True)

  automated_stopping_config: Optional[
      automated_stopping.AutomatedStoppingConfig] = attr.field(
          init=True,
          default=None,
          validator=attr.validators.optional(
              attr.validators.instance_of(
                  automated_stopping.AutomatedStoppingConfig)),
          on_setattr=attr.setters.validate,
          kw_only=True)

  metadata: common.Metadata = attr.field(
      init=True,
      kw_only=True,
      factory=common.Metadata,
      validator=attr.validators.instance_of(common.Metadata),
      on_setattr=[attr.setters.convert, attr.setters.validate])

  # An internal representation as a StudyConfig proto.
  # If this object was created from a StudyConfig proto, a copy of the original
  # proto is kept, to make sure that unknown proto fields are preserved in
  # round trip serialization.
  # TODO: Fix the broken proto validation.
  _study_config: study_pb2.StudySpec = attr.field(
      init=True,
      factory=study_pb2.StudySpec,
      kw_only=True)

  # Public attributes, methods and properties.
  @classmethod
  def from_proto(cls, proto: study_pb2.StudySpec) -> 'StudyConfig':
    """Converts a StudyConfig proto to a StudyConfig object.

    Args:
      proto: StudyConfig proto.

    Returns:
      A StudyConfig object.
    """
    metric_information = base_study_config.MetricsConfig(
        sorted([
            proto_converters.MetricInformationConverter.from_proto(m)
            for m in proto.metrics
        ],
               key=lambda x: x.name))

    oneof_name = proto.WhichOneof('automated_stopping_spec')
    if not oneof_name:
      automated_stopping_config = None
    else:
      automated_stopping_config = automated_stopping.AutomatedStoppingConfig.from_proto(
          getattr(proto, oneof_name))

    metadata = common.Metadata()
    for kv in proto.metadata:
      metadata.abs_ns(common.Namespace.decode(kv.ns))[kv.key] = (
          kv.proto if kv.HasField('proto') else kv.value)

    return cls(
        search_space=proto_converters.SearchSpaceConverter.from_proto(proto),
        algorithm=Algorithm(proto.algorithm),
        metric_information=metric_information,
        observation_noise=ObservationNoise(proto.observation_noise),
        automated_stopping_config=automated_stopping_config,
        study_config=copy.deepcopy(proto),
        metadata=metadata)

  def to_proto(self) -> study_pb2.StudySpec:
    """Serializes this object to a StudyConfig proto."""
    proto = copy.deepcopy(self._study_config)
    proto.algorithm = self.algorithm.value
    proto.observation_noise = self.observation_noise.value

    del proto.metrics[:]
    proto.metrics.extend(
        proto_converters.MetricsConfigConverter.to_proto(
            self.metric_information))

    del proto.parameters[:]
    proto.parameters.extend(
        proto_converters.SearchSpaceConverter.parameter_protos(
            self.search_space))

    if self.automated_stopping_config is not None:
      auto_stop_proto = self.automated_stopping_config.to_proto()
      if isinstance(auto_stop_proto,
                    study_pb2.StudySpec.DecayCurveAutomatedStoppingSpec):
        proto.decay_curve_stopping_spec.CopyFrom(auto_stop_proto)
      elif isinstance(auto_stop_proto,
                      study_pb2.StudySpec.DecayCurveAutomatedStoppingSpec):
        proto.median_automated_stopping_spec.CopyFrom(auto_stop_proto)

    for ns in self.metadata.namespaces():
      ns_string = ns.encode()
      ns_layer = self.metadata.abs_ns(ns)
      for key, value in ns_layer.items():
        metadata_util.assign(proto, key=key, ns=ns_string, value=value)
    return proto

  @property
  def is_single_objective(self) -> bool:
    """Returns True if only one objective metric is configured."""
    return len(self.metric_information) == 1

  @property
  def single_objective_metric_name(self) -> Optional[str]:
    """Returns the name of the single-objective metric, if set.

    Returns:
      String: name of the single-objective metric.
      None: if this is not a single-objective study.
    """
    if len(self.metric_information) == 1:
      return list(self.metric_information)[0].name
    return None

  def _trial_to_external_values(
      self, pytrial: trial.Trial) -> Dict[str, Union[float, int, str, bool]]:
    """Returns the trial paremeter values cast to external types."""
    parameter_values: Dict[str, Union[float, int, str]] = {}
    external_values: Dict[str, Union[float, int, str, bool]] = {}
    # parameter_configs is a list of Tuple[parent_name, ParameterConfig].
    parameter_configs: List[Tuple[Optional[str],
                                  parameter_config.ParameterConfig]] = [
                                      (None, p)
                                      for p in self.search_space.parameters
                                  ]
    remaining_parameters = copy.deepcopy(pytrial.parameters)
    # Traverse the conditional tree using a BFS.
    while parameter_configs and remaining_parameters:
      parent_name, pc = parameter_configs.pop(0)
      parameter_configs.extend(
          (pc.name, child) for child in pc.child_parameter_configs)
      if pc.name not in remaining_parameters:
        continue
      if parent_name is not None:
        # This is a child parameter. If the parent was not seen,
        # skip this parameter config.
        if parent_name not in parameter_values:
          continue
        parent_value = parameter_values[parent_name]
        if parent_value not in pc.matching_parent_values:
          continue
      parameter_values[pc.name] = remaining_parameters[pc.name].value
      if pc.external_type is None:
        external_value = remaining_parameters[pc.name].value
      else:
        external_value = remaining_parameters[pc.name].cast(pc.external_type)  # pytype: disable=wrong-arg-types
      external_values[pc.name] = external_value
      remaining_parameters.pop(pc.name)
    return external_values

  def trial_parameters(
      self, proto: study_pb2.Trial) -> Dict[str, ParameterValueSequence]:
    """Returns the trial values, cast to external types, if they exist.

    Args:
      proto:

    Returns:
      Parameter values dict: cast to each parameter's external_type, if exists.
      NOTE that the values in the dict may be a Sequence as opposed to a single
      element.

    Raises:
      ValueError: If the trial parameters do not exist in this search space.
      ValueError: If the trial contains duplicate parameters.
    """
    pytrial = proto_converters.TrialConverter.from_proto(proto)
    return self._pytrial_parameters(pytrial)

  def _pytrial_parameters(
      self, pytrial: trial.Trial) -> Dict[str, ParameterValueSequence]:
    """Returns the trial values, cast to external types, if they exist.

    Args:
      pytrial:

    Returns:
      Parameter values dict: cast to each parameter's external_type, if exists.
      NOTE that the values in the dict may be a Sequence as opposed to a single
      element.

    Raises:
      ValueError: If the trial parameters do not exist in this search space.
      ValueError: If the trial contains duplicate parameters.
    """
    trial_external_values: Dict[str, Union[float, int, str, bool]] = (
        self._trial_to_external_values(pytrial))
    if len(trial_external_values) != len(pytrial.parameters):
      raise ValueError('Invalid trial for this search space: failed to convert '
                       'all trial parameters: {}'.format(pytrial))

    # Combine multi-dimensional parameter values to a list of values.
    trial_final_values: Dict[str, ParameterValueSequence] = {}
    # multi_dim_params: Dict[str, List[Tuple[int, ParameterValueSequence]]]
    multi_dim_params = collections.defaultdict(list)
    for name in trial_external_values:
      base_index = SearchSpaceSelector.parse_multi_dimensional_parameter_name(
          name)
      if base_index is None:
        trial_final_values[name] = trial_external_values[name]
      else:
        base_name, index = base_index
        multi_dim_params[base_name].append((index, trial_external_values[name]))
    for name in multi_dim_params:
      multi_dim_params[name].sort(key=lambda x: x[0])
      trial_final_values[name] = [x[1] for x in multi_dim_params[name]]

    return trial_final_values

  def trial_metrics(self,
                    proto: study_pb2.Trial,
                    *,
                    include_all_metrics=False) -> Dict[str, float]:
    """Returns the trial's final measurement metric values.

    If the trial is not completed, or infeasible, no metrics are returned.
    By default, only metrics configured in the StudyConfig are returned
    (e.g. only objective and safety metrics).

    Args:
      proto:
      include_all_metrics: If True, all metrics in the final measurements are
        returned. If False, only metrics configured in the StudyConfig are
        returned.

    Returns:
      Dict[metric name, metric value]
    """
    pytrial = proto_converters.TrialConverter.from_proto(proto)
    return self._pytrial_metrics(
        pytrial, include_all_metrics=include_all_metrics)

  def _pytrial_metrics(self,
                       pytrial: trial.Trial,
                       *,
                       include_all_metrics=False) -> Dict[str, float]:
    """Returns the trial's final measurement metric values.

    If the trial is not completed, or infeasible, no metrics are returned.
    By default, only metrics configured in the StudyConfig are returned
    (e.g. only objective and safety metrics).

    Args:
      pytrial:
      include_all_metrics: If True, all metrics in the final measurements are
        returned. If False, only metrics configured in the StudyConfig are
        returned.

    Returns:
      Dict[metric name, metric value]
    """
    configured_metrics = [m.name for m in self.metric_information]

    metrics: Dict[str, float] = {}
    if pytrial.is_completed and not pytrial.infeasible:
      for name in pytrial.final_measurement.metrics:
        if (include_all_metrics or
            (not include_all_metrics and name in configured_metrics)):
          # Special case: Measurement always adds an empty metric by default.
          # If there is a named single objective in study_config, drop the empty
          # metric.
          if not name and self.single_objective_metric_name != name:
            continue
          metrics[name] = pytrial.final_measurement.metrics[name].value
    return metrics



"""Common classes shared between Study and Trial."""

import collections
from collections import abc
from typing import DefaultDict, Dict, overload
from typing import Iterable, List, Optional, Tuple, TypeVar, Union, Type
import attr

from google.protobuf import any_pb2
from google.protobuf.message import Message

M = TypeVar('M', bound=Message)
T = TypeVar('T')
MetadataValue = Union[str, any_pb2.Any, Message]

# Namespace Encoding.
#
# By definition, ∀ ns ∈ Namespace, Namespace.decode(ns.encode()) == ns.
# The tricky part of that definition is handling namespaces with components
# that are empty strings.  Notably, we want to make sure that
# Namespace(()).encode() != Namespace(('',)).encode().
# So, we set up the mapping:
# Namespace(()).encode() -> ''
# Namespace((s,)).encode() -> ':s'
# Namespace((s, s)).encode() -> ':s:s',
# et cetera, and note that every tuple gets a unique encoding, even if $s is the
# empty string.  (As long as we escape colons properly.)
#
# So, ns.encode() is a bijection, therefore it has an inverse which we call
# Namespace.decode(s).


def _parse(arg: str) -> Tuple[str, ...]:
  """Parses an encoded namespace string into a namespace tuple."""
  # The tricky part here is that arg.split('') has a length of 1, so it can't
  # generate a zero-length tuple; we handle that corner case manually.
  if not arg:
    return ()
  # And, then, once we've handled the case of _parse(''), we note that all the
  # other encoded strings begin with a colon.  It thus contains no information
  # and we can remove it.
  # TODO: Once we're on Python 3.9, use: arg = arg.removeprefix(':')
  if arg.startswith(':'):
    arg = arg[1:]
  # The rest of the algorithm is that we split on all colons, both
  # escaped and unescaped.  Then, we walk through the list of fragments and
  # join back together the colons that were preceeded by an escape character,
  # dropping the escape character as we go.
  fragments = arg.split(':')
  output = []
  join = False
  for frag in fragments:
    if join and frag and frag[-1] == '\\':
      output[-1] += ':' + frag[:-1]
      join = True
    elif join:  # Doesn't end in an escape character.
      output[-1] += ':' + frag
      join = False
    elif frag and frag[-1] == '\\':  # Don't join to previous.
      output.append(frag[:-1])
      join = True
    else:  # Don't join to previous and doesn't end in an escape.
      output.append(frag)
      join = False
  return tuple(output)


@attr.frozen(eq=True, order=True, hash=True, auto_attribs=True, init=False)
class Namespace(abc.Sequence):
  r"""A namespace for the Metadata class.

  Namespaces form a tree; a particular namespace can be thought of as a tuple of
  namespace components.

  You can create a Namespace from a string, with Namespace.decode(s), where
  the string is parsed into components, splitting at colons; decode('a:b') gives
  you a two-component namespace: ('a', 'b').
  Or, you can create that same Namespace from a tuple of strings/components
  e.g. by constructing Namespace(('a', 'b')).  In the tuple case, the strings
  are not parsed and colons are ordinary characters.

  TLDR: If you decode() a namespace from a string, then ":" is a
    reserved character, but when constructing from a tuple, there are no
    reserved characters.

  Decoding the string form:
  * Initial colons don't matter: Namespace.decode(':a') == Namespace('a');
    this is a single-component namespace.
  * Colons separate components:
    Namespace.decode('a:b') == Namespace(['a', 'b']).
    (This is a two-component namespace.)
  * Colons are encoded as r'\:':
    Namespace.decode('a\\:b') == Namespace(('a:b')).
    (This is a single-component namespace.)

  Conversions: For a Namespace x,
  * Namespace.decode(x.encode()) == x; here, x.encode() will be a string with
    colons separating the components.
  * Namespaces act as a Sequence[str], so Namespace(tuple(x)) == x and
    Namespace(x) == x.
  """

  _as_tuple: Tuple[str, ...] = attr.field(hash=True, eq=True, order=True)

  def __init__(self, arg: Iterable[str] = ()):
    """Generates a Namespace from its component strings.

    Args:
      arg: a tuple representation of a namespace.
    """
    arg = tuple(arg)
    self.__attrs_init__(as_tuple=arg)

  _ns_repr_table = str.maketrans({':': r'\:'})

  @classmethod
  def decode(cls, s: str) -> 'Namespace':
    r"""Decode a string into a Namespace.

    For a Namespace x, Namespace.decode(x.encode()) == x.

    Args:
      s: A string where ':' separates namespace components, and colon is
        escaped as r'\:'.

    Returns:
      A namespace.
    """
    return Namespace(_parse(s))

  def encode(self) -> str:
    """Encodes a Namespace into a string.

    Given a Namespace x, Namespace.decode(x.encode()) == x.

    Returns:
      Colons are escaped, then Namespace components are joined by colons.
    """
    return ''.join(
        [':' + c.translate(self._ns_repr_table) for c in self._as_tuple])

  def __len__(self) -> int:
    """Number of components (elements of the tuple form)."""
    return len(self._as_tuple)

  def __add__(self, other: Iterable[str]) -> 'Namespace':
    """Appends components onto the namespace."""
    return Namespace(self._as_tuple + tuple(other))

  @overload
  def __getitem__(self, key: int) -> str:
    ...

  @overload
  def __getitem__(self, key: slice) -> 'Namespace':
    ...

  def __getitem__(self, key):
    """Retrieves item by the specified key."""
    if isinstance(key, int):
      return self._as_tuple[key]
    return Namespace(self._as_tuple[key])

  def __str__(self) -> str:
    """Shows the namespace, fully escaped."""
    return self.encode()

  def __repr__(self) -> str:
    """Shows the namespace, fully escaped."""
    return f'Namespace({self.encode()})'

  def startswith(self, prefix: Iterable[str]) -> bool:
    """Returns True if this namespace starts with prefix."""
    ns_prefix = Namespace(prefix)
    return self[:len(ns_prefix)] == ns_prefix


class _MetadataSingleNameSpace(Dict[str, MetadataValue]):
  """Stores metadata associated with one namespace."""
  pass


class Metadata(abc.MutableMapping):
  """Metadata class.

  This is the main interface for reading metadata from a Trial (writing metadata
  should typically be done via the MetadataUpdateContext class.)

  This behaves like a str->str dict, within a given namespace.
    mm = Metadata({'foo': 'Foo'})
    mm.get('foo')  # Returns 'Foo'
    mm['foo']      # Returns 'Foo'
    mm['bar'] = 'Bar'
    mm.update({'a': 'A'}, gleep='Gleep')

  1. Keys are namespaced. Each Metadata object only interacts with one
    Namespace, but a metadata object and its children share a
    common set of (namespace, key, value) triplets.

    Namespaces form a tree, and you can walk down the tree.  There are two
    namespace operators: ns(s) which adds component(s) on to the namespace, and
    abs_ns() which specifies the entire namespace.

    A Metadata() object is always created at the root of the namespace tree,
    and the root is special (it's the only namespace that Vizier users can write
    or conveniently read).  Pythia algorithm developers should avoid the root
    namespace, unless they intend to pass data to/from Vizier users.

    mm = Metadata({'foo': 'foofoo'})
    mm.ns('NewName')['bar'] = 'Bar'
    mm['foo']               # Returns 'foofoo'
    mm['bar']               # Throws a KeyError
    mm.ns('NewName')['foo'] # Throws a KeyError
    mm.ns('NewName')['bar'] # Returns 'Bar'
    mm.ns('NewName').get('bar') # Returns 'Bar'

    # Use of abs_ns().
    mm = Metadata()
    mm.abs_ns(Namespace(('NewName',)))['bar'] = 'Bar'
    mm.abs_ns(Namespace(('NewName',)))  # returns 'Bar'

    # Multi-component namespaces.
    mm = Metadata()
    mm.ns('a').ns('b')['foo'] = 'AB-foo'
    mm.ns('a')['foo'] = 'A-foo'
    mm['foo']          # Throws a KeyError
    mm.ns('a')['foo']  # returns 'A-foo'
    mm.ns('a').ns('b')['foo']  # returns 'AB-foo'
    mm.abs_ns(Namespace(('a', 'b'))).get('foo')  # Returns 'ab-foo'
    mm.abs_ns(Namespace.decode('a:b')).get('foo')  # Returns 'ab-foo'

  2. Values can be protobufs. If `metadata['foo']` is an instance of `MyProto`
    proto message or `Any` proto that packs a `MyProto` message, then the proto
    can be recovered by calling:
      my_proto = metadata.get_proto('foo', cls=MyProto)
      isinstance(my_proto, MyProto) # Returns `True`

  3. An iteration over a Metadata object only shows you the data in the current
    namespace.  So,

    mm = Metadata({'foo': 'foofoo'})
    for k, v in mm.ns('gleep'):
      ...

    will not yield anything because there are no keys in the 'gleep' namespace.
    WARNING: Because of this behavior, Metadata(mm) will quietly drop metadata
      from all but mm's current namespace.

    Be aware that type(v) is MetadataValue, which includes protos in addition to
    strings.

    To iterate over all the keys in all the namespaces use the namespaces()
    method.

    mm : Metadata
    for ns in mm.namespaces():
      for k, v in mm.abs_ns(ns).items():
        ...
  """

  def __init__(self, *args: Union[Dict[str, MetadataValue],
                                  Iterable[Tuple[str, MetadataValue]]],
               **kwargs: MetadataValue):
    """Construct; this follows dict(), and puts data in the root namespace.

    You can pass it a dict, or an object that yields (key, value)
    pairs, and those pairs will be put in the root namespace.

    Args:
      *args: A dict or an iterable the yields key-value pairs.
      **kwargs: key=value pairs to be added to the specified namespace.
    """
    self._stores: DefaultDict[
        Namespace, _MetadataSingleNameSpace] = collections.defaultdict(
            _MetadataSingleNameSpace)
    self._namespace = Namespace()
    self._store = self._stores[self._namespace]
    self._store.update(*args, **kwargs)

  def abs_ns(self, namespace: Iterable[str] = ()) -> 'Metadata':
    """Switches to a specified absolute namespace.

    All the Metadata object's data is shared between $self and the returned
    object, but the new Metadata object will have a different default
    namespace.

    Args:
      namespace: a list of Namespace components.  (Defaults to the root, empty
      Namespace.)

    Returns:
      A new Metadata object in the specified namespace; the new object shares
      data (except the namespace) with $self.
    """
    return self._copy_core(Namespace(namespace))

  def ns(self, component: str) -> 'Metadata':
    r"""Switches to a deeper namespace by appending a component.

    All the metadata is shared between $self and the returned value, but they
    have a different current namespace.

    Args:
      component: one component to be added to the current namespace.

    Returns:
      A new Metadata object in the specified namespace; the new object shares
      metadata (except the choice of namespace) with $self.
    """
    new_ns: Namespace = self._namespace + (component,)
    return self._copy_core(new_ns)

  def __repr__(self) -> str:
    itemlist: List[str] = []
    for namespace, store in self._stores.items():
      item_string = f'(namespace:{namespace}, items: {store})'
      itemlist.append(item_string)
    return 'Metadata({}, current_namespace={})'.format(', '.join(itemlist),
                                                       self._namespace.encode())

  def __str__(self) -> str:
    return 'namespace: {} items: {}'.format(str(self._namespace), self._store)

  def get_proto(self, key: str, *, cls: Type[M]) -> Optional[M]:
    """Deprecated.

    Use get() instead.

    Gets the metadata as type `cls`, or None if not possible.

    Args:
      key:
      cls: Pass in a proto ***class***, not a proto object.

    Returns:
      Proto message, if the value associated with the key exists and
      can be parsed into cls; None otherwise.
    """
    value = self._store.get(key, None)
    if value is None:
      return None

    if isinstance(value, cls):
      # Starting from 3.10, pytype supports typeguard, which obsoletes
      # the need for the `pytype:disable` clause.
      return value  # pytype: disable=bad-return-type

    if isinstance(value, any_pb2.Any):
      # `value` is an Any proto potentially packing `cls`.
      message = cls()
      success = value.Unpack(message)
      return message if success else None

    return None

  def get(self,
          key: str,
          default: Optional[T] = None,
          *,
          cls: Type[T] = str) -> Optional[T]:
    """Gets the metadata as type `cls`, or None if not possible.

    Given regular string values, this function behaves exactly like a
    regular string-to-string dict (within its namespace).
      metadata = common.Metadata({'key': 'value'})
      assert metadata.get('key') == 'value'
      assert metadata.get('badkey', 'badvalue') == 'badvalue'

    Example with numeric string values:
      metadata = common.Metadata({'float': '1.2', 'int': '60'})
      assert metadata.get('float', cls=float) == 1.2
      assert metadata.get('badkey', 0.2, cls=float) == 0.2
      assert metadata.get('int', cls=int) == 60
      assert metadata.get('badkey', 1, cls=int) == 1

    Example with `Duration` and `Any` proto values:
      duration = Duration(seconds=60)
      anyproto = Any()
      anyproto.Pack(duration)
      metadata = common.Metadata({'duration': duration, 'any': anyproto})
      assert metadata.get('duration', cls=Duration) == duration
      assert metadata.get('any', cls=Duration) == duration

    Args:
      key:
      default: Default value.
      cls: Desired type of the value.

    Returns:
      Default if the key does not exist. Otherwise, the matching value is
      parsed into type `cls`. For proto messages, it involves unpacking
      Any proto.
    """
    try:
      value = self._store[key]
    except KeyError:
      return default
    if isinstance(value, cls):
      # Starting from 3.10, pytype supports typeguard, which obsoletes
      # the need for the `pytype:disable` clause.
      return value  # pytype: disable=bad-return-type
    if isinstance(value, any_pb2.Any):
      # `value` is an Any proto potentially packing `cls`.
      message = cls()
      success = value.Unpack(message)
      return message if success else None
    return cls(value)

  # TODO: Rename to `abs_namespaces`
  def namespaces(self) -> Tuple[Namespace, ...]:
    """Get all namespaces for which there is at least one key.

    Returns:
      For all `ns` in `md.namespaces()`, `md.abs_ns(ns)` is not empty.
    """
    return tuple([ns for ns, store in self._stores.items() if store])

  # TODO: Rename to `namespaces`
  def subnamespaces(self) -> Tuple[Namespace, ...]:
    """Returns relative namespaces that are at or below the current namespace.

    For all `ns` in `md.subnamespaces()`, `md.abs_ns(md.current_ns() + ns)` is
    not empty.  E.g. if namespace 'foo:bar' is non-empty, and you're in
    namespace 'foo', then the result will contain namespace 'bar'.

    Returns:
      For namespaces that begin with the current namespace and are
      non-empty, this returns a namespace object that contains the relative
      path from the current namespace.
    """
    return tuple([
        Namespace(ns[len(self._namespace):])
        for ns, store in self._stores.items()
        if store and ns.startswith(self._namespace)
    ])

  def current_ns(self) -> Namespace:
    """Displays the object's current Namespace."""
    return self._namespace

  # START OF abstract methods inherited from `MutableMapping` base class.
  def __getitem__(self, key: str) -> MetadataValue:
    return self._store.__getitem__(key)

  def __setitem__(self, key: str, value: MetadataValue):
    self._store[key] = value

  def __delitem__(self, key: str):
    del self._store[key]

  def __iter__(self):
    return iter(self._store)

  def __len__(self):
    return len(self._store)

  def __copy__(self) -> 'Metadata':
    """Shallow copy -- metadata continues to be shared.

    Returns:
      A copy of the object.
    """
    return self._copy_core(self._namespace)

  # END OF Abstract methods inherited from `MutableMapping` base class.

  def _copy_core(self, ns: Namespace) -> 'Metadata':
    """Shallow copy: metadata is shared, default namespace changes.

    Args:
      ns: the namespace to use for the new object.

    Returns:
      A copy of the object.
    """
    md = Metadata()
    md._namespace = ns  # pylint: disable='protected-access'
    md._stores = self._stores  # pylint: disable='protected-access'
    md._store = md._stores[md._namespace]  # pylint: disable='protected-access'
    return md

  def update(self, *args: Union[Dict[str, MetadataValue],
                                Iterable[Tuple[str, MetadataValue]]],
             **kwargs: MetadataValue) -> None:
    self._store.update(*args, **kwargs)

  def attach(self, other: 'Metadata') -> None:
    """Attach the $other metadata as a descendent of this metadata.

    More precisely, it takes the part of `other`'s namespace that is at or
    below `other`'s current namespace, and attaches it to `self`'s current
    namespace.
    * Tree structure is preserved and nothing is flattened.
    * Attached data overwrites existing data, item-by-item, not
      namepace-by-namespace.

    So, if we have
    other = Metadata()
    other.abs_ns(Namespace.(('x', 'y', 'z'))['foo'] = 'bar'
    m = Metadata()
    m.ns('w').attach(other.ns('x'))
    then
    m.abs_ns(('w', 'y', 'z'))['foo'] will contain 'bar'.

    Args:
      other: a Metadata object to copy from.
    """
    for ns in other.subnamespaces():
      self._stores[self._namespace + ns].update(
          other.abs_ns(other.current_ns() + ns))



import collections
from collections import abc as collections_abc
import copy
import enum
import math
import re
from typing import Callable, Iterable, Iterator, List, Optional, Sequence, Tuple, Type, TypeVar, Union, overload

from absl import logging
import attr
import numpy as np
from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import parameter_config
from vizier._src.pyvizier.shared import trial

################### PyTypes ###################
ScaleType = parameter_config.ScaleType
ExternalType = parameter_config.ExternalType
# A sequence of possible internal parameter values.
MonotypeParameterSequence = parameter_config.MonotypeParameterSequence
_T = TypeVar('_T')


################### Helper Classes ###################
def _min_leq_max(instance: 'MetricInformation', _, value: float):
  if value > instance.max_value:
    raise ValueError(
        f'min_value={value} cannot exceed max_value={instance.max_value}.')


def _max_geq_min(instance: 'MetricInformation', _, value: float):
  if value < instance.min_value:
    raise ValueError(
        f'min_value={instance.min_value} cannot exceed max_value={value}.')


# Values should NEVER be removed from ObjectiveMetricGoal, only added.
class ObjectiveMetricGoal(enum.IntEnum):
  """Valid Values for MetricInformation.Goal."""
  MAXIMIZE = 1
  MINIMIZE = 2

  # pylint: disable=comparison-with-callable
  @property
  def is_maximize(self) -> bool:
    return self == self.MAXIMIZE

  @property
  def is_minimize(self) -> bool:
    return self == self.MINIMIZE


class MetricType(enum.Enum):
  """Type of the metric.

  OBJECTIVE: Objective to be maximized / minimized.
  SAFETY: Objective to be kept above / below a certain threshold.
  """
  OBJECTIVE = 'OBJECTIVE'
  SAFETY = 'SAFETY'  # Soft constraint

  # pylint: disable=comparison-with-callable
  @property
  def is_safety(self) -> bool:
    return self == MetricType.SAFETY

  @property
  def is_objective(self) -> bool:
    return self == MetricType.OBJECTIVE


@attr.define(frozen=False, init=True, slots=True)
class MetricInformation:
  """MetricInformation provides optimization metrics configuration."""

  # The name of this metric. An empty string is allowed for single-metric
  # optimizations.
  name: str = attr.field(
      init=True, default='', validator=attr.validators.instance_of(str))

  goal: ObjectiveMetricGoal = attr.field(
      init=True,
      # pylint: disable=g-long-lambda
      converter=ObjectiveMetricGoal,
      validator=attr.validators.instance_of(ObjectiveMetricGoal),
      on_setattr=[attr.setters.convert, attr.setters.validate],
      kw_only=True)

  # The following are only valid for Safety metrics.
  # safety_threshold should always be set to a float (default 0.0), for safety
  # metrics.
  safety_threshold: Optional[float] = attr.field(
      init=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(float)),
      kw_only=True)
  safety_std_threshold: Optional[float] = attr.field(
      init=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(float)),
      kw_only=True)
  percentage_unsafe_trials_threshold: Optional[float] = attr.field(
      init=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(float)),
      kw_only=True)

  # Minimum value of this metric can be optionally specified.
  min_value: float = attr.field(
      init=True,
      default=None,
      # FYI: Converter is applied before validator.
      converter=lambda x: float(x) if x is not None else -np.inf,
      validator=[attr.validators.instance_of(float), _min_leq_max],
      kw_only=True)

  # Maximum value of this metric can be optionally specified.
  max_value: float = attr.field(
      init=True,
      default=None,
      # FYI: Converter is applied before validator.
      converter=lambda x: float(x) if x is not None else np.inf,
      validator=[attr.validators.instance_of(float), _max_geq_min],
      on_setattr=attr.setters.validate,
      kw_only=True)

  def min_value_or(self, default_value_fn: Callable[[], float]) -> float:
    """Returns the minimum value if finite, or default_value_fn().

    Avoids the common pitfalls of using
      `metric.min_value or default_value`
    which would incorrectly use the default_value when min_value == 0, and
    requires default_value to have been computed.

    Args:
      default_value_fn: Default value if min_value is not finite.
    """
    if np.isfinite(self.min_value):
      return self.min_value
    else:
      return default_value_fn()

  def max_value_or(self, default_value_fn: Callable[[], float]) -> float:
    """Returns the minimum value if finite, or default_value_fn().

    Avoids the common pitfalls of using
      `metric.max_value or default_value`
    which would incorrectly use the default_value when max_value == 0, and
    requires default_value to have been computed.

    Args:
      default_value_fn: Default value if max_value is not configured.
    """
    if np.isfinite(self.max_value):
      return self.max_value
    else:
      return default_value_fn()

  @property
  def range(self) -> float:
    """Range of the metric. Can be infinite."""
    return self.max_value - self.min_value

  @property
  def type(self) -> MetricType:
    if (self.safety_threshold is not None or
        self.safety_std_threshold is not None):
      return MetricType.SAFETY
    else:
      return MetricType.OBJECTIVE

  def flip_goal(self) -> 'MetricInformation':
    """Flips the goal in-place and returns the reference to self."""
    if self.goal == ObjectiveMetricGoal.MAXIMIZE:
      self.goal = ObjectiveMetricGoal.MINIMIZE
    else:
      self.goal = ObjectiveMetricGoal.MAXIMIZE
    return self


@attr.define(frozen=False, init=True, slots=True)
class MetricsConfig(collections_abc.Collection):
  """Container for metrics.

  Metric names should be unique.
  """
  _metrics: List[MetricInformation] = attr.ib(
      init=True,
      factory=list,
      converter=list,
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of(MetricInformation),
          iterable_validator=attr.validators.instance_of(Iterable)))

  def item(self) -> MetricInformation:
    if len(self._metrics) != 1:
      raise ValueError('Can be called only when there is exactly one metric!')
    return self._metrics[0]

  def _assert_names_are_unique(self) -> None:
    counts = collections.Counter(metric.name for metric in self._metrics)
    if len(counts) != len(self._metrics):
      for name, count in counts.items():
        if count > 1:
          raise ValueError(f'Duplicate metric name: {name} in {self._metrics}')

  def __attrs_post_init__(self):
    self._assert_names_are_unique()

  def __iter__(self) -> Iterator[MetricInformation]:
    return iter(self._metrics)

  def __contains__(self, x: object) -> bool:
    return x in self._metrics

  def __len__(self) -> int:
    return len(self._metrics)

  def __add__(self, metrics: Iterable[MetricInformation]) -> 'MetricsConfig':
    return MetricsConfig(self._metrics + list(metrics))

  def of_type(
      self, include: Union[MetricType,
                           Iterable[MetricType]]) -> 'MetricsConfig':
    """Filters the Metrics by type."""
    if isinstance(include, MetricType):
      include = (include,)
    return MetricsConfig(m for m in self._metrics if m.type in include)

  def append(self, metric: MetricInformation):
    self._metrics.append(metric)
    self._assert_names_are_unique()

  def extend(self, metrics: Iterable[MetricInformation]):
    for metric in metrics:
      self.append(metric)

  @property
  def is_single_objective(self) -> bool:
    """Returns True if only one objective metric is configured."""
    return len(self.of_type(MetricType.OBJECTIVE)) == 1


@attr.s(frozen=True, init=True, slots=True, kw_only=True)
class _PathSegment:
  """Selection of a parameter name and one of its values."""
  # A ParameterConfig name.
  name: str = attr.ib(
      init=True, validator=attr.validators.instance_of(str), kw_only=True)

  # A ParameterConfig value.
  value: Union[int, float, str] = attr.ib(
      init=True,
      validator=attr.validators.instance_of((int, float, str)),
      kw_only=True)


class _PathSelector(Sequence[_PathSegment]):
  """Immutable sequence of path segments."""

  def __init__(self, iterable: Iterable[_PathSegment] = tuple()):
    self._paths = tuple(iterable)

  @overload
  def __getitem__(self, s: slice) -> '_PathSelector':
    ...

  @overload
  def __getitem__(self, i: int) -> _PathSegment:
    ...

  def __getitem__(self, index):
    item = self._paths[index]
    if isinstance(item, _PathSegment):
      return item
    else:
      return _PathSelector(item)

  def __len__(self) -> int:
    """Returns the number of elements in the container."""
    return len(self._paths)

  def __add__(
      self, other: Union[Sequence[_PathSegment],
                         _PathSegment]) -> '_PathSelector':
    if isinstance(other, _PathSegment):
      other = [other]
    return _PathSelector(self._paths + tuple(other))

  def __repr__(self) -> str:
    """Returns the path as a string."""
    path = '/'.join(['{}={}'.format(p.name, p.value) for p in self._paths])
    return f'{path}'


class InvalidParameterError(Exception):
  """Error thrown when parameter values are invalid."""


################### Main Classes ###################
@attr.s(frozen=True, init=True, slots=True, kw_only=True)
class SearchSpaceSelector:
  """A Selector for all, or part of a SearchSpace."""

  # List of ParameterConfig objects referenced by this selector.
  # This is a reference to a list of objects owned by SearchSpace (and will
  # typically include the entire SearchSpace).
  _configs: List[parameter_config.ParameterConfig] = attr.ib(
      init=True,
      factory=list,
      # Verify that this is a list of ParameterConfig objects.
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of(
              parameter_config.ParameterConfig),
          iterable_validator=attr.validators.instance_of(list)),
      kw_only=True)

  # _selected_path and _selected_name control how parameters are added to the
  # search space.
  #
  # 1) If _selected_path is empty, and _selected_name is empty, parameters
  #    are added to the root of the search space.
  # 2) If _selected_path is empty, and _selected_name is non-empty, parameters
  #    will be added as child parameters to all root and child parameters
  #    with name ==_selected_name.
  # 3) If both _selected_path and _selected_name are specified, parameters will
  #    be added as child parameters to the parameter specified by the path and
  #    the name.
  # 4) If _selected_path is non-empty, and _selected_name is empty, this is an
  #    error.

  # An ordered list of _PathSelector objects which uniquely identifies a path
  # in a conditional tree.
  _selected_path: _PathSelector = attr.ib(
      init=True,
      default=_PathSelector(),
      converter=_PathSelector,
      # Verify that this is a list of _PathSegment objects.
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of(_PathSegment),
          iterable_validator=attr.validators.instance_of(Iterable)),
      kw_only=True)

  # A ParameterConfig name.
  # If there is a _selected_name, then there have to also be _selected_values
  # below, and new parameters are added to the parent(s) selected by
  # _selected_path and _selected_name.
  _selected_name: str = attr.ib(
      init=True,
      default='',
      validator=attr.validators.instance_of(str),
      kw_only=True)

  # List of ParameterConfig values from _configs.
  # If there are _selected_values, then there have to also be _selected_name
  # above.
  _selected_values: MonotypeParameterSequence = attr.ib(
      init=True,
      factory=list,
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of((int, float, str)),
          iterable_validator=attr.validators.instance_of(list)),
      kw_only=True)

  @property
  def parameter_name(self) -> str:
    """Returns the selected parameter name."""
    return self._selected_name

  @property
  def parameter_values(self) -> MonotypeParameterSequence:
    """Returns the selected parameter values."""
    return copy.deepcopy(self._selected_values)

  def add_float_param(self,
                      name: str,
                      min_value: float,
                      max_value: float,
                      *,
                      default_value: Optional[float] = None,
                      scale_type: Optional[ScaleType] = ScaleType.LINEAR,
                      index: Optional[int] = None) -> 'SearchSpaceSelector':
    """Adds floating point parameter config(s) to the search space.

    If select_all() was previously called for this selector, so it contains
    selected parent values, the parameter configs will be added as child
    parameters to the selected parameter configs, and a reference to this
    selector is returned.

    If no parent values are selected, the parameter config(s) will be added at
    the same level as currently selected parameters, and a reference to the
    newly added parameters is returned.

    Args:
      name: The parameter's name. Cannot be empty.
      min_value: Inclusive lower bound for the parameter.
      max_value: Inclusive upper bound for the parameter.
      default_value: A default value for the Parameter.
      scale_type: Scaling to be applied. NOT VALIDATED.
      index: Specifies the multi-dimensional index for this parameter. E.g. if
        name='rate' and index=0, then a single ParameterConfig with name
        'rate[0]' is added. `index` should be >= 0.

    Returns:
      SearchSpaceSelector(s) for the newly added parameter(s):
      One SearchSpaceSelector if one parameter was added, or a list of
      SearchSpaceSelector if multiple parameters were added.

    Raises:
      ValueError: If `index` is invalid (e.g. negative).
    """
    bounds = (float(min_value), float(max_value))
    param_names = self._get_parameter_names_to_create(name=name, index=index)

    new_params = []
    for param_name in param_names:
      new_pc = parameter_config.ParameterConfig.factory(
          name=param_name,
          bounds=bounds,
          scale_type=scale_type,
          default_value=default_value)
      new_params.append(new_pc)
    return self._add_parameters(new_params)[0]

  def add_int_param(self,
                    name: str,
                    min_value: int,
                    max_value: int,
                    *,
                    default_value: Optional[int] = None,
                    scale_type: Optional[ScaleType] = None,
                    index: Optional[int] = None) -> 'SearchSpaceSelector':
    """Adds integer parameter config(s) to the search space.

    If select_all() was previously called for this selector, so it contains
    selected parent values, the parameter configs will be added as child
    parameters to the selected parameter configs, and a reference to this
    selector is returned.

    If no parent values are selected, the parameter config(s) will be added at
    the same level as currently selected parameters, and a reference to the
    newly added parameters is returned.

    Args:
      name: The parameter's name. Cannot be empty.
      min_value: Inclusive lower bound for the parameter.
      max_value: Inclusive upper bound for the parameter.
      default_value: A default value for the Parameter.
      scale_type: Scaling to be applied. NOT VALIDATED.
      index: Specifies the multi-dimensional index for this parameter. E.g. if
        name='hidden_units' and index=0, then a single ParameterConfig with name
        'hidden_units[0]' is added. `index` should be >= 0.

    Returns:
      SearchSpaceSelector for the newly added parameter.

    Raises:
      ValueError: If min_value or max_value are not integers.
      ValueError: If `index` is invalid (e.g. negative).
    """
    int_min_value = int(min_value)
    if not math.isclose(min_value, int_min_value):
      raise ValueError('min_value for an INTEGER parameter should be an integer'
                       ', got: [{}]'.format(min_value))
    int_max_value = int(max_value)
    if not math.isclose(max_value, int_max_value):
      raise ValueError('max_value for an INTEGER parameter should be an integer'
                       ', got: [{}]'.format(min_value))
    bounds = (int_min_value, int_max_value)

    param_names = self._get_parameter_names_to_create(name=name, index=index)

    new_params = []
    for param_name in param_names:
      new_pc = parameter_config.ParameterConfig.factory(
          name=param_name,
          bounds=bounds,
          scale_type=scale_type,
          default_value=default_value)
      new_params.append(new_pc)
    return self._add_parameters(new_params)[0]

  def add_discrete_param(
      self,
      name: str,
      feasible_values: Union[Sequence[float], Sequence[int]],
      *,
      default_value: Optional[Union[float, int]] = None,
      scale_type: Optional[ScaleType] = ScaleType.LINEAR,
      index: Optional[int] = None,
      auto_cast: Optional[bool] = True) -> 'SearchSpaceSelector':
    """Adds ordered numeric parameter config(s) with a finite set of values.

    IMPORTANT: If a parameter is discrete, its values are assumed to have
    ordered semantics. Thus, you should not use discrete parameters for
    unordered values such as ids. In this case, see add_categorical_param()
    below.

    If select_all() was previously called for this selector, so it contains
    selected parent values, the parameter configs will be added as child
    parameters to the selected parameter configs, and a reference to this
    selector is returned.

    If no parent values are selected, the parameter config(s) will be added at
    the same level as currently selected parameters, and a reference to the
    newly added parameters is returned.

    Args:
      name: The parameter's name. Cannot be empty.
      feasible_values: The set of feasible values for this parameter.
      default_value: A default value for the Parameter.
      scale_type: Scaling to be applied. NOT VALIDATED.
      index: Specifies the multi-dimensional index for this parameter. E.g. if
        name='batch_size' and index=0, then a single ParameterConfig with name
        'batch_size[0]' is added. `index` should be >= 0.
      auto_cast: If False, the external type will be set to INTEGER if all
        values are castable to an integer without losing precision. If True, the
        external type will be set to float.

    Returns:
      SearchSpaceSelector for the newly added parameter.

    Raises:
      ValueError: If `index` is invalid (e.g. negative).
    """
    param_names = self._get_parameter_names_to_create(name=name, index=index)

    external_type = ExternalType.FLOAT
    if auto_cast:
      # If all feasible values are convertible to ints without loss of
      # precision, annotate the external type as INTEGER. This will cast
      # [0., 1., 2.] into [0, 1, 2] when parameter values are returned in
      # clients.
      if all([v == round(v) for v in feasible_values]):
        external_type = ExternalType.INTEGER

    new_params = []
    for param_name in param_names:
      new_pc = parameter_config.ParameterConfig.factory(
          name=param_name,
          feasible_values=sorted(feasible_values),
          scale_type=scale_type,
          default_value=default_value,
          external_type=external_type)
      new_params.append(new_pc)
    return self._add_parameters(new_params)[0]

  def add_categorical_param(
      self,
      name: str,
      feasible_values: Sequence[str],
      *,
      default_value: Optional[str] = None,
      scale_type: Optional[ScaleType] = None,
      index: Optional[int] = None) -> 'SearchSpaceSelector':
    """Adds unordered string-valued parameter config(s) to the search space.

    IMPORTANT: If a parameter is categorical, its values are assumed to be
    unordered. If the `feasible_values` have ordering, use add_discrete_param()
    above, since it will improve Vizier's model quality.

    If select_all() was previously called for this selector, so it contains
    selected parent values, the parameter configs will be added as child
    parameters to the selected parameter configs, and a reference to this
    selector is returned.

    If no parent values are selected, the parameter config(s) will be added at
    the same level as currently selected parameters, and a reference to the
    newly added parameters is returned.

    Args:
      name: The parameter's name. Cannot be empty.
      feasible_values: The set of feasible values for this parameter.
      default_value: A default value for the Parameter.
      scale_type: Scaling to be applied. NOT VALIDATED.
      index: Specifies the multi-dimensional index for this parameter. E.g. if
        name='id' and index=0, then a single ParameterConfig with name 'id[0]'
        is added. `index` should be >= 0.

    Returns:
      SearchSpaceSelector for the newly added parameter.

    Raises:
      ValueError: If `index` is invalid (e.g. negative).
    """
    param_names = self._get_parameter_names_to_create(name=name, index=index)

    new_params = []
    for param_name in param_names:
      new_pc = parameter_config.ParameterConfig.factory(
          name=param_name,
          feasible_values=sorted(feasible_values),
          scale_type=scale_type,
          default_value=default_value)
      new_params.append(new_pc)
    return self._add_parameters(new_params)[0]

  def add_bool_param(self,
                     name: str,
                     feasible_values: Optional[Sequence[bool]] = None,
                     *,
                     default_value: Optional[bool] = None,
                     scale_type: Optional[ScaleType] = None,
                     index: Optional[int] = None) -> 'SearchSpaceSelector':
    """Adds boolean-valued parameter config(s) to the search space.

    If select_all() was previously called for this selector, so it contains
    selected parent values, the parameter configs will be added as child
    parameters to the selected parameter configs, and a reference to this
    selector is returned.

    If no parent values are selected, the parameter config(s) will be added at
    the same level as currently selected parameters, and a reference to the
    newly added parameters is returned.

    Args:
      name: The parameter's name. Cannot be empty.
      feasible_values: An optional list of feasible boolean values, i.e. one of
        the following: [True], [False], [True, False], [False, True].
      default_value: A default value for the Parameter.
      scale_type: Scaling to be applied. NOT VALIDATED.
      index: Specifies the multi-dimensional index for this parameter. E.g. if
        name='match' and index=0, then a single ParameterConfig with name
        'match[0]' is added. `index` should be >= 0.

    Returns:
      SearchSpaceSelector for the newly added parameter.

    Raises:
      ValueError: If `feasible_values` has invalid values.
      ValueError: If `index` is invalid (e.g. negative).
    """
    allowed_values = (None, (True, False), (False, True), (True,), (False,))
    if feasible_values not in allowed_values:
      raise ValueError('feasible_values must be one of %s; got: %s.' %
                       (allowed_values, feasible_values))
    # Boolean parameters are represented as categorical parameters internally.
    bool_to_string = lambda x: 'True' if x else 'False'
    if feasible_values is None:
      categories = ('True', 'False')
    else:
      categories = [bool_to_string(x) for x in feasible_values]
    feasible_values = sorted(categories, reverse=True)

    if default_value is not None:
      default_value = bool_to_string(default_value)

    param_names = self._get_parameter_names_to_create(name=name, index=index)

    new_params = []
    for param_name in param_names:
      new_pc = parameter_config.ParameterConfig.factory(
          name=param_name,
          feasible_values=sorted(feasible_values),
          scale_type=scale_type,
          default_value=default_value,
          external_type=ExternalType.BOOLEAN)
      new_params.append(new_pc)
    return self._add_parameters(new_params)[0]

  def select(
      self,
      parameter_name: str,
      parameter_values: Optional[MonotypeParameterSequence] = None
  ) -> 'SearchSpaceSelector':
    """Selects a single parameter specified by path and parameter_name.

    This method should be called to select a parent parameter, before calling
    `add_*_param` methods to create child parameters.

    Given a selector to the root of the search space:
    root = pyvizier.SearchSpace().select_root()

    1) To select a parameter at the root of the search space, with parent values
      for child parameters:
      model = root.select('model_type', ['dnn'])
      model.add_float_param('hidden_units', ...)
    2) To select a parameter at the root of the search space, and defer parent
      value selection to later calls:
      model = root.select('model_type')
      # Add `hidden_units` and `optimizer_type` as `dnn` children.
      model.select_values(['dnn']).add_float_param('hidden_units', ...)
      model.select_values(['dnn']).add_categorical_param(
        'optimizer_type', ['adam', 'adagrad'])
      # Add `optimizer_type` and `activation` as `linear` children.
      model.select_values(['linear']).add_categorical_param(
        'optimizer_type', ['adam', 'ftrl'])
      model.select_values(['linear']).add_categorical_param('activation', ...)
    3) To select a parameter in a conditional search space, specify a path, by
      chaining select() calls:
      optimizer = root.select('model_type', ['linear']).select('optimizer_type')
      optimizer.select_values('adam').add_float_param('learning_rate', 0.001,..)
      optimizer.select_values('ftrl').add_float_param('learning_rate', 0.1,..)

      # OR pre-select the parent parameter values:
      optimizer = root.select('model_type', ['linear']).select(
        'optimizer_type', ['adam'])
      optimizer.add_float_param('learning_rate', 0.001,...)
    4) If there is *only one* parameter with the given name, then it is possible
      to select it without specifying the path, using:
      selectors = root.select_all('activation')
      # 'activation' exists only under model_type='linear'.
      assert len(selectors) == 1
      activation = selectors[0]

    Args:
      parameter_name:
      parameter_values: Optional parameter values for this selector, which will
        be used to add child parameters, or traverse a conditional tree.

    Returns:
      A new SearchSpaceSelector.
    """
    # Make sure parameter_name exists in the conditional parameters tree.
    # parameter_values will be validated only when a child parameter is added.
    if not self._parameter_exists(parameter_name):
      raise ValueError('No parameter with name {} exists in this SearchSpace')

    path = []
    selected_values = []
    if parameter_values is not None:
      if not isinstance(parameter_values, (list, tuple)):
        raise ValueError('parameter_values should be a list or tuple, given '
                         '{} with type {}'.format(parameter_values,
                                                  type(parameter_values)))
      selected_values = parameter_values

    if self._selected_name:
      # There is already a parameter name selected, so this is a chained select
      # call.
      if not self._selected_values:
        raise ValueError('Cannot call select() again before parameter values '
                         'are selected: parameter {} was previously selected, '
                         ' with the path: {}, but no values were selected for '
                         'it'.format(self.parameter_name, self.path_string))
      # Return a new selector, with the currently selected parameter added to
      # the path.
      new_path_segment = [
          _PathSegment(
              name=self._selected_name, value=self._selected_values[0])
      ]
      path = self._selected_path + new_path_segment
      if not self._path_exists(path):
        raise ValueError('Path {} does not exist in this SearchSpace: '
                         '{}'.format((path), self))

    return SearchSpaceSelector(
        configs=self._configs,
        selected_path=path,
        selected_name=parameter_name,
        selected_values=selected_values)

  def select_values(
      self,
      parameter_values: MonotypeParameterSequence) -> 'SearchSpaceSelector':
    """Selects values for a pre-selected parameter.

    This method should be called to select parent parameter(s) value(s), before
    calling `add_*_param` methods to create child parameters.

    This method must be called AFTER select().
    This method mutates this selector.

    Args:
      parameter_values: Parameter values for this selector, which will be used
        to add child parameters.

    Returns:
      SearchSpaceSelector
    """
    if not self._selected_name:
      raise ValueError('No parameter is selected. Call select() first.')
    if not parameter_values:
      raise ValueError(
          'parameter_values cannot be empty. Specify at least one value.')
    if not isinstance(parameter_values, (list, tuple)):
      raise ValueError('parameter_values should be a list or tuple, given '
                       '{} with type {}'.format(parameter_values,
                                                type(parameter_values)))
    # TODO: Allow to directly select boolean parent parameters.
    object.__setattr__(self, '_selected_values', parameter_values)
    return self

  def select_all(
      self, parameter_name: str, parameter_values: MonotypeParameterSequence
  ) -> List['SearchSpaceSelector']:
    """Select one or more parent parameters, with the same name.

    This method should be called to select parent parameter(s), before calling
    `add_*_param` methods to create child parameters.
    Multiple parent parameters with the same name are possible in a conditional
    search space. See go/conditional-parameters for more details.

    1) If the conditional search space has two parameters with the same
    name, 'optimizer_type', given a selector to the root of the search space,
    select_all() can be used to simultaneously add child parameters to both
    'optimizer_type` parameters:

    root = pyvizier.SearchSpace().select_root()
    model.select_values(['dnn']).add_categorical_param(
        'optimizer_type', ['adam', 'adagrad'])
    model.select_values(['linear']).add_categorical_param(
        'optimizer_type', ['adam', 'ftrl'])
    # Add a 'learning_rate' parameter to both 'adam' optimizers:
    optimizers = model.select_all('optimizer_type', parent_values=['adam'])
    optimizers.add_float_param('learning_rate', ...)

    2) If there is *only one* parameter with the given name, then it is also
      possible to use select_all() to select it:
      root = pyvizier.SearchSpace().select_root()
      model.select_values(['dnn']).add_categorical_param('activation', ...)
      # Select the single parameter with the name 'activation':
      selectors = root.select_all('activation')
      assert len(selectors) == 1
      activation = selector[0]

    Args:
      parameter_name:
      parameter_values: Optional parameter values for this selector, which will
        be used to add child parameters.

    Returns:
      List of SearchSpaceSelector
    """
    # TODO: Raise an error if this selector already has selected_name.
    # Make sure parameter_name exists in the conditional parameters tree.
    if not self._parameter_exists(parameter_name):
      raise ValueError('No parameter with name {} exists in this SearchSpace')

    if parameter_values is not None:
      if not isinstance(parameter_values, (list, tuple)):
        raise ValueError('parameter_values should be a list or tuple, given '
                         '{} with type {}'.format(parameter_values,
                                                  type(parameter_values)))
    # TODO: Complete this method.
    raise NotImplementedError()

  def _path_exists(self, path: _PathSelector) -> bool:
    """Checks if the path exists in the conditional tree."""
    for parent in self._configs:
      if (path[0].name == parent.name and
          path[0].value in parent.feasible_values):
        if len(path) == 1:
          # No need to recurse.
          return True
        return self._path_exists_inner(parent, path[1:])
    return False

  @classmethod
  def _path_exists_inner(cls, current_root: parameter_config.ParameterConfig,
                         current_path: _PathSelector) -> bool:
    """Returns true if the path exists, starting at root_parameter."""
    child_idx = None
    for idx, child in enumerate(current_root.child_parameter_configs):
      if (current_path[0].name == child.name and
          current_path[0].value in child.feasible_values):
        child_idx = idx
        break
    if child_idx is None:
      # No match is found. This path does not exist.
      return False
    if len(current_path) == 1:
      # This is the end of the path.
      return True
    # Keep traversing.
    return cls._path_exists_inner(
        current_root.child_parameter_configs[child_idx], current_path[1:])

  def _parameter_exists(self, parameter_name: str) -> bool:
    """Checks if there exists at least one parameter with this name.

    Note that this method checks existence in the entire search space.

    Args:
      parameter_name:

    Returns:
      bool: Exists.
    """
    found = False
    for parent in self._configs:
      for pc in parent.traverse(show_children=False):
        if pc.name == parameter_name:
          found = True
          break
    return found

  @classmethod
  def _get_parameter_names_to_create(cls,
                                     *,
                                     name: str,
                                     length: Optional[int] = None,
                                     index: Optional[int] = None) -> List[str]:
    """Returns the names of all parameters which should be created.

    Args:
      name: The base parameter name.
      length: Specifies the length of a multi-dimensional parameters. If larger
        than 1, then multiple ParameterConfigs are added. E.g. if name='rate'
        and length=2, then two ParameterConfigs with names 'rate[0]', 'rate[1]'
        are added. Cannot be specified together with `index`.
      index: Specifies the multi-dimensional index for this parameter. Cannot be
        specified together with `length`. E.g. if name='rate' and index=1, then
        a single ParameterConfig with name 'rate[1]' is added.

    Returns:
      List of parameter names to create.

    Raises:
      ValueError: If `length` or `index` are invalid.
    """
    if length is not None and index is not None:
      raise ValueError('Only one of `length` and `index` can be specified. Got'
                       ' length={}, index={}'.format(length, index))
    if length is not None and length < 1:
      raise ValueError('length must be >= 1. Got length={}'.format(length))
    if index is not None and index < 0:
      raise ValueError('index must be >= 0. Got index={}'.format(index))

    param_names = []
    if length is None and index is None:
      # Add one parameter with no multi-dimensional index.
      param_names.append(name)
    elif index is not None:
      # Add one parameter with a multi-dimensional index.
      param_names.append(cls._multi_dimensional_parameter_name(name, index))
    elif length is not None:
      # `length > 0' is synthatic sugar for multi multi-dimensional parameter.
      # Each multi-dimensional parameter is encoded as a list of separate
      # parameters with names equal to `name[index]` (index is zero based).
      for i in range(length):
        param_names.append(cls._multi_dimensional_parameter_name(name, i))
    return param_names

  @classmethod
  def _multi_dimensional_parameter_name(cls, name: str, index: int) -> str:
    """Returns the indexed parameter name."""
    return '{}[{}]'.format(name, index)

  @classmethod
  def parse_multi_dimensional_parameter_name(
      cls, name: str) -> Optional[Tuple[str, int]]:
    """Returns the base name for a multi-dimensional parameter name.

    Args:
      name: A parameter name.

    Returns:
      (base_name, index): if name='hidden_units[10]', base_name='hidden_units'
        and index=10.
      Returns None if name is not in the format 'base_name[idx]'.
    """
    regex = r'(?P<name>[^()]*)\[(?P<index>\d+)\]$'
    pattern = re.compile(regex)
    matches = pattern.match(name)
    if matches is None:
      return None
    return (matches.groupdict()['name'], int(matches.groupdict()['index']))

  @property
  def path_string(self) -> str:
    """Returns the selected path as a string."""
    return str(self._selected_path)

  def _add_parameters(
      self, parameters: List[parameter_config.ParameterConfig]
  ) -> List['SearchSpaceSelector']:
    """Adds ParameterConfigs either to the root, or as child parameters.

    Args:
      parameters: The parameters to add to the search space.

    Returns:
      A list of SearchSpaceSelectors, one for each parameters added.
    """
    if self._selected_name and not self._selected_values:
      raise ValueError(
          'Cannot add child parameters to parameter {}: parent values were '
          'not selected. Call select_values() first.'.format(
              self._selected_name))
    if not self._selected_name and self._selected_values:
      raise ValueError(
          'Cannot add child parameters: no parent name is selected.'
          ' Call select() or select_all() first.')
    if self._selected_path and not self._selected_name:
      raise ValueError(
          'Cannot add child parameters: path is specified ({}), but no parent'
          ' name is specified. Call select() or select_all() first'.format(
              self.path_string))

    selectors: List['SearchSpaceSelector'] = []
    if not self._selected_path and not self._selected_name:
      # If _selected_path is empty, and _selected_name is empty, parameters
      # are added to the root of the search space.
      logging.info('Adding parameters to root of search space: %s', parameters)
      self._configs.extend(parameters)
      # Return Selectors for the newly added parameters.
      for param in parameters:
        selectors.append(
            SearchSpaceSelector(
                configs=self._configs,
                selected_path=[],
                selected_name=param.name,
                selected_values=[]))
    elif not self._selected_path and self._selected_name:
      # If _selected_path is empty, and _selected_name is not empty, parameters
      # will be added as child parameters to *all* root and child parameters
      # with name ==_selected_name.
      logging.info(
          'Adding child parameters to all matching parents with '
          'name "%s": %s', self._selected_name, parameters)
      for idx, root_param in enumerate(self._configs):
        updated_param, new_selectors = self._recursive_add_child_parameters(
            self._configs, _PathSelector(), root_param, self._selected_name,
            self._selected_values, parameters)
        # Update the root ParameterConfig in place.
        self._configs[idx] = updated_param
        selectors.extend(new_selectors)
    else:
      # If both _selected_path and _selected_name are specified, parameters will
      # be added as child parameters to the parameter specified by the path and
      # the name.
      logging.info(
          'Adding child parameters to parent with path "%s" '
          ', name "%s", values: %s: %s', (self._selected_path),
          self._selected_name, self._selected_values, parameters)
      idx, updated_param, new_selectors = self._add_parameters_at_selected_path(
          root_configs=self._configs,
          complete_path=self._selected_path,
          parent_name=self._selected_name,
          parent_values=self._selected_values,
          new_children=parameters)
      # Update the root ParameterConfig in place.
      self._configs[idx] = updated_param
      selectors.extend(new_selectors)

    if not selectors:
      raise ValueError(
          'Cannot add child parameters: the path ({}), is not valid.'.format(
              self.path_string))
    return selectors

  @classmethod
  def _recursive_add_child_parameters(
      cls, configs: List[parameter_config.ParameterConfig], path: _PathSelector,
      root: parameter_config.ParameterConfig, parent_name: str,
      parent_values: MonotypeParameterSequence,
      new_children: List[parameter_config.ParameterConfig]
  ) -> Tuple[parameter_config.ParameterConfig, List['SearchSpaceSelector']]:
    """Recursively adds new children to all matching parameters.

    new_children are potentially added to root, and all matching child
    parameters with name==parent_name.

    Args:
      configs: A list of configs to include in returned SearchSpaceSelectors,
        this list is not modified or used for anything else.
      path: The path to include in returned SearchSpaceSelectors.
      root: Parent parameter to start the recursion at.
      parent_name: new_children are added to all parameter with this name.
      parent_values: new_children are added with these parent values.
      new_children: Child parameter configs to add.

    Returns:
      (An updated root with all of its children updated, list of selectors to
       any parameters which may have been added)
    """
    updated_children: List[Tuple[MonotypeParameterSequence,
                                 parameter_config.ParameterConfig]] = []
    selectors: List['SearchSpaceSelector'] = []

    logging.debug(
        '_recursive_add_child_parameters called with path=%s, '
        'root=%s, parent_name=%s, parent_values=%s', path, root, parent_name,
        parent_values)

    if root.name == parent_name:
      # Add new children to this root. If this is a leaf parameter,
      # e.g. it has no children, this is where the recursion ends.
      for child in new_children:
        updated_children.append((parent_values, child))
        # For the path, select one parent value, since for the path, the exact
        # value does not matter, as long as it's valid.
        root_path_fragment = [
            _PathSegment(name=root.name, value=parent_values[0])
        ]
        selectors.append(
            SearchSpaceSelector(
                configs=configs,
                selected_path=path + root_path_fragment,
                selected_name=child.name,
                selected_values=[]))
    # Recursively update existing children, if any.
    for child in root.child_parameter_configs:
      # For the path, select one parent value, since for the path, the exact
      # value does not matter, as long as it's valid.
      root_path_fragment = [
          _PathSegment(name=root.name, value=child.matching_parent_values[0])
      ]
      updated_child, new_selectors = cls._recursive_add_child_parameters(
          configs, path + root_path_fragment, child, parent_name, parent_values,
          new_children)
      updated_children.append(
          (updated_child.matching_parent_values, updated_child))
      selectors += new_selectors
    # Update all children (existing and potentially new) in the root.
    return root.clone_without_children.add_children(updated_children), selectors

  @classmethod
  def _add_parameters_at_selected_path(
      cls, root_configs: List[parameter_config.ParameterConfig],
      complete_path: _PathSelector, parent_name: str,
      parent_values: MonotypeParameterSequence,
      new_children: List[parameter_config.ParameterConfig]
  ) -> Tuple[int, parameter_config.ParameterConfig,
             List['SearchSpaceSelector']]:
    """Adds new children to the parameter specified by the path and parent_name.

    Args:
      root_configs: A list of configs to include in returned
        SearchSpaceSelectors, this list is not modified. These are expected to
        be the configs at the root of the search space.
      complete_path: The path to include in the returned SearchSpaceSelectors.
      parent_name: new_children are added to all parameter with this name.
      parent_values: new_children are added with these parent values.
      new_children: Child parameter configs to add.

    Returns:
      (Root index in root_configs,
       an updated root with all of its children updated,
       list of selectors to any parameters which may have been added)

    Raises:
      RuntimeError:
      ValueError:
    """
    if not complete_path:
      # This is an internal error, since the caller should never specify an
      # empty current_path.
      raise RuntimeError('Internal error: got empty complete_path')

    # This is the beginning of the recursion. Select a root to recurse at.
    current_root: Optional[parameter_config.ParameterConfig] = None
    root_idx: int = 0
    for root_idx, root_param in enumerate(root_configs):
      if complete_path[0].name == root_param.name:
        current_root = root_param
        break
    if current_root is None:
      raise ValueError('Invalid path: {}: failed to traverse the path: failed'
                       ' to find a matching root for parameter name "{}".'
                       ' Root parameter names: {}'.format(
                           (complete_path), complete_path[0].name,
                           [pc.name for pc in root_configs]))

    updated_root, selectors = cls._add_parameters_at_selected_path_inner(
        root_configs=root_configs,
        complete_path=complete_path,
        current_root=current_root,
        current_path=complete_path[1:],
        parent_name=parent_name,
        parent_values=parent_values,
        new_children=new_children)
    return (root_idx, updated_root, selectors)

  @classmethod
  def _add_parameters_at_selected_path_inner(
      cls, root_configs: List[parameter_config.ParameterConfig],
      complete_path: _PathSelector,
      current_root: parameter_config.ParameterConfig,
      current_path: _PathSelector, parent_name: str,
      parent_values: MonotypeParameterSequence,
      new_children: List[parameter_config.ParameterConfig]
  ) -> Tuple[parameter_config.ParameterConfig, List['SearchSpaceSelector']]:
    """Adds new children to the parameter specified by the path and parent_name.

    Args:
      root_configs: A list of configs to include in returned
        SearchSpaceSelectors, this list is not modified. These are expected to
        be the configs at the root of the search space.
      complete_path: The path to include in the returned SearchSpaceSelectors.
      current_root: Parent parameter to start the recursion at.
      current_path: The path to the parent parameter from current_root. This is
        used in the recursion.
      parent_name: new_children are added to all parameter with this name.
      parent_values: new_children are added with these parent values.
      new_children: Child parameter configs to add.

    Returns:
      (An updated root with all of its children updated,
       List of selectors to all added parameters)

    Raises:
      RuntimeError:
      ValueError:
    """
    updated_children: List[Tuple[MonotypeParameterSequence,
                                 parameter_config.ParameterConfig]] = []
    selectors: List['SearchSpaceSelector'] = []

    if not current_path:
      # This is the end of the path. End the recursion.
      # parent_name should be a child of current_root
      child_idx = None
      for idx, child in enumerate(current_root.child_parameter_configs):
        if parent_name == child.name:
          child_idx = idx
          last_parent_path = [
              _PathSegment(name=parent_name, value=parent_values[0])
          ]
          new_path = complete_path + last_parent_path
          updated_child, selectors = cls._add_child_parameters(
              root_configs, new_path, child, parent_values, new_children)
          break
      if child_idx is None:
        raise ValueError('Invalid parent_name: after traversing the path "{}", '
                         'failed to find a child parameter with name "{}".'
                         ' Current root="{}"'.format((complete_path),
                                                     parent_name, current_root))

      # Update current_root with the updated child.
      for idx, child in enumerate(current_root.child_parameter_configs):
        if idx == child_idx:
          updated_children.append(
              (updated_child.matching_parent_values, updated_child))
        else:
          updated_children.append((child.matching_parent_values, child))
      return (
          current_root.clone_without_children.add_children(updated_children),
          selectors)

    # Traverse the path: find which child matches the next path selection.
    child_idx = None
    for idx, child in enumerate(current_root.child_parameter_configs):
      if (current_path[0].name == child.name and
          current_path[0].value in child.feasible_values):
        child_idx = idx
        break
    if child_idx is None:
      raise ValueError('Invalid path: "{}": failed to traverse the path: failed'
                       ' to find a matching child for path selector "{}".'
                       ' Current root="{}", current_path="{}"'.format(
                           (complete_path), (current_path[:1]),
                           current_root.name, (current_path)))

    updated_child, selectors = cls._add_parameters_at_selected_path_inner(
        root_configs=root_configs,
        complete_path=complete_path,
        current_root=current_root.child_parameter_configs[child_idx],
        current_path=current_path[1:],
        parent_name=parent_name,
        parent_values=parent_values,
        new_children=new_children)
    # Update current_root with the updated child, leave the selectors untouched.
    for idx, child in enumerate(current_root.child_parameter_configs):
      if idx == child_idx:
        updated_children.append(
            (updated_child.matching_parent_values, updated_child))
      else:
        updated_children.append((child.matching_parent_values, child))
    return (current_root.clone_without_children.add_children(updated_children),
            selectors)

  @classmethod
  def _add_child_parameters(
      cls, selector_configs: List[parameter_config.ParameterConfig],
      selector_path: _PathSelector, parent: parameter_config.ParameterConfig,
      parent_values: MonotypeParameterSequence,
      new_children: List[parameter_config.ParameterConfig]
  ) -> Tuple[parameter_config.ParameterConfig, List['SearchSpaceSelector']]:
    """Adds new children to the parent parameter and returns selectors.

    Args:
      selector_configs: A list of configs to include in returned
        SearchSpaceSelectors, this list is not modified. These are expected to
        be the configs at the root of the search space.
      selector_path: The path to include in the returned SearchSpaceSelectors.
      parent: Parent parameter to add children to.
      parent_values: new_children are added with these parent values.
      new_children: Child parameter configs to add.

    Returns:
      (An updated root with all of its children updated,
       List of selectors to all added parameters)

    Raises:
      RuntimeError:
      ValueError:
    """
    updated_children: List[Tuple[MonotypeParameterSequence,
                                 parameter_config.ParameterConfig]] = []
    selectors: List['SearchSpaceSelector'] = []

    # Add existing children.
    for child in parent.child_parameter_configs:
      updated_children.append((child.matching_parent_values, child))
    # Add new child parameter configs.
    for child in new_children:
      updated_children.append((parent_values, child))
      selectors.append(
          SearchSpaceSelector(
              configs=selector_configs,
              selected_path=selector_path,
              selected_name=child.name,
              selected_values=[]))
    # Add all children (existing and potentially new) to the parent.
    return (parent.clone_without_children.add_children(updated_children),
            selectors)


@attr.s(frozen=True, init=True, slots=True, kw_only=True)
class SearchSpace:
  """A builder and wrapper for StudyConfig.parameter_configs."""

  _parameter_configs: List[parameter_config.ParameterConfig] = attr.ib(
      init=False, factory=list)

  @classmethod
  def _factory(
      cls: Type[_T],
      parameter_configs: Optional[List[parameter_config.ParameterConfig]] = None
  ) -> _T:
    """Creates a new SearchSpace containing the provided parameter configs.

    Args:
      parameter_configs:

    Returns:
      SearchSpace
    """
    if parameter_configs is None:
      parameter_configs = []
    space = cls()
    object.__setattr__(space, '_parameter_configs', list(parameter_configs))
    return space

  @property
  def parameters(self) -> List[parameter_config.ParameterConfig]:
    """Returns COPIES of the parameter configs in this Space."""
    return copy.deepcopy(self._parameter_configs)

  def select_root(self) -> SearchSpaceSelector:
    """Returns a selector for the root of the search space.

    Parameters can be added to the search space using the returned
    SearchSpaceSelector.
    """
    return SearchSpaceSelector(configs=self._parameter_configs)

  @property
  def is_conditional(self) -> bool:
    """Returns True if search_space contains any conditional parameters."""
    return any([p.child_parameter_configs for p in self._parameter_configs])

  def contains(self, parameters: trial.ParameterDict) -> bool:
    try:
      self.assert_contains(parameters)
      return True
    except InvalidParameterError:
      return False

  def assert_contains(self, parameters: trial.ParameterDict) -> bool:
    """Throws an error if parameters is not a valid point in the space.

    Args:
      parameters:

    Returns:
      Always returns True unless an exception is Raised.

    Raises:
      InvalidParameterError: If parameters are invalid.
      NotImplementedError: If parameter type is unknown
    """
    if self.is_conditional:
      raise NotImplementedError('Not implemented for conditional space.')
    if len(parameters) != len(self._parameter_configs):
      set1 = set(pc.name for pc in self._parameter_configs)
      set2 = set(parameters)
      raise InvalidParameterError(
          f'Search space has {len(self._parameter_configs)} parameters '
          f'but only {len(parameters)} were given. '
          f'Missing in search space: {set2 - set1}. '
          f'Missing in parameters: {set1 - set2}.')
    for pc in self._parameter_configs:
      if pc.name not in parameters:
        raise InvalidParameterError(f'{pc.name} is missing in {parameters}.')
      elif not pc.contains(parameters[pc.name]):
        raise InvalidParameterError(
            f'{parameters[pc.name]} is not feasible in {pc}')
    return True


################### Main Class ###################
@attr.define(frozen=False, init=True, slots=True)
class ProblemStatement:
  """A builder and wrapper for core StudyConfig functionality."""

  search_space: SearchSpace = attr.ib(
      init=True,
      factory=SearchSpace,
      validator=attr.validators.instance_of(SearchSpace))

  metric_information: MetricsConfig = attr.ib(
      init=True,
      factory=MetricsConfig,
      converter=MetricsConfig,
      validator=attr.validators.instance_of(MetricsConfig),
      kw_only=True)

  metadata: common.Metadata = attr.field(
      init=True,
      kw_only=True,
      factory=common.Metadata,
      validator=attr.validators.instance_of(common.Metadata))

  @property
  def debug_info(self) -> str:
    return ''

  @classmethod
  def from_problem(cls: Type[_T], problem: 'ProblemStatement') -> _T:
    return cls(
        search_space=problem.search_space,
        metric_information=problem.metric_information,
        metadata=problem.metadata)

  def to_problem(self) -> 'ProblemStatement':
    return ProblemStatement(
        search_space=self.search_space,
        metric_information=self.metric_information,
        metadata=self.metadata)



"""Policy supporters that keep data in RAM."""
import datetime
from typing import Iterable, List, Optional, Sequence

import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.pythia import policy
from vizier._src.pythia import policy_supporter
from vizier.pyvizier import converters
from vizier.pyvizier import multimetric


# TODO: Keep the Pareto frontier trials.
@attr.s(frozen=True, init=True, slots=True)
class LocalPolicyRunner(policy_supporter.PolicySupporter):
  """Runs a fresh Study in RAM using a Policy.

  LocalPolicyRunner acts as a limited vizier service + client that runs in RAM.
  Trials can only be added and never removed.

  Example of using a policy to run a Study for 100 iterations, 1 trial each:
    runner = LocalPolicyRunner(my_study_config)
    policy = MyPolicy(runner)
    for _ in range(100):
      trials = runner.SuggestTrials(policy, count=1)
      if not trials:
        break
      for t in trials:
        t.complete(vz.Measurement(
            {'my_objective': my_objective(t)}, inplace=True))

  Attributes:
    study_config: Study config.
    study_guid: Unique identifier for the study.
  """

  study_config: vz.StudyConfig = attr.ib(
      init=True, validator=attr.validators.instance_of(vz.StudyConfig))
  study_guid: str = attr.ib(init=True, kw_only=True, default='')
  _trials: List[vz.Trial] = attr.ib(init=False, factory=list)

  @property
  def trials(self) -> Sequence[vz.Trial]:
    return self._trials

  def study_descriptor(self) -> vz.StudyDescriptor:
    return vz.StudyDescriptor(
        self.study_config, guid=self.study_guid, max_trial_id=len(self._trials))

  def _check_study_guid(self, study_guid: Optional[str]) -> None:
    if study_guid is not None and self.study_guid != study_guid:
      raise ValueError('LocalPolicyRunner does not support accessing '
                       'other studies than the current one, which has '
                       'guid={self.study_guid}')

  def GetStudyConfig(self, study_guid: Optional[str] = None) -> vz.StudyConfig:
    self._check_study_guid(study_guid)
    return self.study_config

  def GetTrials(
      self,
      *,
      study_guid: Optional[str] = None,
      trial_ids: Optional[Iterable[int]] = None,
      min_trial_id: Optional[int] = None,
      max_trial_id: Optional[int] = None,
      status_matches: Optional[vz.TrialStatus] = None,
      include_intermediate_measurements: bool = True) -> List[vz.Trial]:
    self._check_study_guid(study_guid)
    min_trial_id = min_trial_id or 1
    max_trial_id = max_trial_id or (len(self._trials))
    trials = [
        t for t in self._trials[min_trial_id - 1:max_trial_id]
        if (status_matches is None or t.status == status_matches)
    ]
    if trial_ids is None:
      return trials
    else:
      trial_ids = set(trial_ids)
      return [t for t in trials if t.id in trial_ids]

  def CheckCancelled(self, note: Optional[str] = None) -> None:
    pass

  def TimeRemaining(self) -> datetime.timedelta:
    return datetime.timedelta(seconds=100.0)

  def SendMetadata(self, delta: policy_supporter.MetadataDelta) -> None:
    for ns in delta.on_study.namespaces():
      self.study_config.metadata.abs_ns(ns).update(delta.on_study.abs_ns(ns))

    for tid, deltum in delta.on_trials.items():
      for ns in deltum.namespaces():
        self._trials[tid - 1].metadata.abs_ns(ns).update(deltum.abs_ns(ns))

  # TODO: Return `count` trials for multi-objectives, when
  # `count` exceeds the size of the pareto frontier.
  def GetBestTrials(self, *, count: Optional[int] = None) -> List[vz.Trial]:
    """Returns optimal trials.

    Single-objective study:
      * If `count` is unset, returns all tied top trials.
      * If `count` is set, returns top `count` trials, breaking ties
           arbitrarily.

    Multi-objective study:
      * If `count` is unset, returns all pareto optimal trials.
      * If `count` is set, returns up to `count` of pareto optimal trials that
          are arbitrarily selected.

    Args:
      count: If unset, returns pareto optimal trials only. If set, returns the
        top "count" trials.

    Returns:
      Best trials.
    """
    if not self.study_config.metric_information.of_type(
        vz.MetricType.OBJECTIVE):
      raise ValueError('Requires at least one objective metric.')
    if self.study_config.metric_information.of_type(vz.MetricType.SAFETY):
      raise ValueError('Cannot work with safe metrics.')

    converter = converters.TrialToArrayConverter.from_study_config(
        self.study_config,
        flip_sign_for_minimization_metrics=True,
        dtype=np.float32)

    if len(self.study_config.metric_information) == 1:
      # Single metric: Sort and take top N.
      count = count or 1  # Defaults to 1.
      labels = converter.to_labels(self._trials).squeeze()
      sorted_idx = np.argsort(-labels)  # np.argsort sorts in ascending order.
      return list(np.asarray(self._trials)[sorted_idx[:count]])
    else:
      algorithm = multimetric.FastParetoOptimalAlgorithm()
      is_optimal = algorithm.is_pareto_optimal(
          points=converter.to_labels(self._trials))
      return list(np.asarray(self._trials)[is_optimal][:count])

  def AddTrials(self, trials: Sequence[vz.Trial]) -> None:
    """(Re-)assigns ids to the trials and add them to the study."""
    for i, trial in enumerate(trials):
      trial.id = i + len(self.trials) + 1
    self._trials.extend(trials)

  def AddSuggestions(
      self, suggestions: Sequence[vz.TrialSuggestion]) -> Sequence[vz.Trial]:
    """Assigns ids to suggestions and add them to the study."""
    trials = []
    for suggestion in suggestions:
      # Assign temporary ids, which will be overwritten by AddTrials() method.
      trials.append(suggestion.to_trial(-1))
    self.AddTrials(trials)
    return trials

  def SuggestTrials(self, algorithm: policy.Policy,
                    count: int) -> Sequence[vz.Trial]:
    """Suggest and add new trials."""
    decisions = algorithm.suggest(
        policy.SuggestRequest(self.study_descriptor(), count))
    return self.AddSuggestions([
        vz.TrialSuggestion(d.parameters, metadata=d.metadata) for d in decisions
    ])



"""Tests for vizier.pyvizier.shared.context."""

from vizier._src.pyvizier.shared import context
from absl.testing import absltest


class ContextTest(absltest.TestCase):

  def testDefaultsNotShared(self):
    """Make sure default parameters are not shared between instances."""
    context1 = context.Context()
    context2 = context.Context()
    context1.parameters['x1'] = context.ParameterValue(5)
    self.assertEmpty(context2.parameters)


if __name__ == '__main__':
  absltest.main()



"""Test runners for algorithms."""
from typing import Any, Optional, Sequence

from absl import logging
import numpy as np

from vizier import algorithms as vza

from vizier import pyvizier as vz


def run_with_random_metrics(
    designer: vza.Designer,
    config: vz.StudyConfig,
    iters: int = 5,
    *,
    batch_size: Optional[int] = 1,
    seed: Any = None,
    verbose: int = 0,
    validate_parameters: bool = False) -> Sequence[vz.Trial]:
  """Generates completed `Trial`s with random metrics.

  EXAMPLE: This method can be used for smoke testing a `Designer`.
  ```
  from vizier._src.algorithms.testing import test_runners
  from vizier.testing import test_studies
  from vizier import pyvizier as vz

  study_config = vz.StudyConfig(
      test_studies.flat_space_with_all_types(),
      [vz.MetricInformation(
          'objective',
          goal=vz.ObjectiveMetricGoal.MAXIMIZE)],
      validate_parameters=True)
  test_runners.run_with_random_metrics(my_designer, study_config)
  ```

  EXAMPLE: This method can be used for generating a large number of trials
  to be used as a test dataset.
  ```
  # (Continued from the above code block)
  from vizier._src.algorithms.designers import random
  trials = test_runners.run_with_random_metrics(
      random.RandomDesigner(study_config.search_space),
      study_config,
      validate_parameters=False)
  ```

  Args:
    designer: Designer object.
    config: Study config.
    iters: Number of suggest-update iterations.
    batch_size: Number of suggestions to ask in each suggest() call.
    seed: Random seed for generating metrics.
    verbose: Increase the verbosity to see more logs.
    validate_parameters: If True, check if the suggested trials are valid in the
      search space.

  Returns:
    This method runs a suggest-update loop after completing suggestions with
    random metric values, and then returns all generated trials.
  """
  rng = np.random.RandomState(seed)
  all_trials = []
  for it in range(iters):
    suggestions = designer.suggest(batch_size)
    if not suggestions:
      logging.info(
          'Preemptively finished at iteration %s'
          'because designer returned nothing.', it)
      break
    trials = []
    for suggestion in designer.suggest(batch_size):
      if validate_parameters:
        config.search_space.assert_contains(suggestion.parameters)
      measurement = vz.Measurement()
      for mi in config.metric_information:
        measurement.metrics[mi.name] = rng.uniform(
            mi.min_value_or(lambda: -10.), mi.max_value_or(lambda: 10.))
      trials.append(suggestion.to_trial(len(trials) + 1).complete(measurement))
      if verbose:
        logging.info('At iteration %s, trials suggested and evaluated:\n%s', it,
                     trials)
      designer.update(vza.CompletedTrials(trials))
    all_trials.extend(trials)

  return all_trials



"""Policy Supporter used for the OSS Vizier service.

The Policy can use these methods to communicate with Vizier.
"""

import datetime
from typing import Iterable, List, Optional
from vizier import pythia
from vizier import pyvizier as vz
from vizier.service import pyvizier
from vizier.service import utils

from vizier.service import vizier_service_pb2
from vizier.service import vizier_service_pb2_grpc


# TODO: Implement UpdateMetadata after metadata additions
class ServicePolicySupporter(pythia.PolicySupporter):
  """Service version of the PolicySupporter."""

  # TODO: Replace vizier_service_instance with a vizier_client.
  def __init__(
      self, study_guid: str,
      vizier_service_instance: vizier_service_pb2_grpc.VizierServiceServicer):
    """Initalization stores a Vizier Service Instance to list Trials.

    Note that this is NOT sending an actual RPC to the server, because the
    policy is running in the same process as the server, but this can be
    trivially modified.

    Args:
      study_guid: A default study_name; the name of this study.
      vizier_service_instance: vizier_service.VizierService() to be used for
        retriving from datastore.
    """
    self._study_guid = study_guid
    self._vizier_service = vizier_service_instance

  def GetStudyConfig(self, study_guid: Optional[str] = None) -> vz.StudyConfig:
    if study_guid is None:
      study_guid = self._study_guid
    request = vizier_service_pb2.GetStudyRequest(name=study_guid)
    study = self._vizier_service.GetStudy(request, None)
    return pyvizier.StudyConfig.from_proto(study.study_spec)

  def GetTrials(
      self,
      *,
      study_guid: Optional[str] = None,  # Equivalent to Study.name in OSS.
      trial_ids: Optional[Iterable[int]] = None,
      min_trial_id: Optional[int] = None,
      max_trial_id: Optional[int] = None,
      status_matches: Optional[vz.TrialStatus] = None,
      include_intermediate_measurements: bool = True) -> List[vz.Trial]:
    """Requests Trials from Vizier."""

    if study_guid is None:
      study_guid = self._study_guid
    request = vizier_service_pb2.ListTrialsRequest(parent=study_guid)
    # Implicitly creates a copy of the data.
    trials = self._vizier_service.ListTrials(request, None).trials
    all_pytrials = pyvizier.TrialConverter.from_protos(trials)

    if trial_ids is not None:
      # We're going to do repeated comparisons in _trial_filter(), so this
      # needs to be more real than an Iterator.
      trial_ids = frozenset(trial_ids)

    def _trial_filter(trial) -> bool:
      if trial_ids is not None and trial.id not in trial_ids:
        return False
      else:
        if min_trial_id is not None:
          if trial.id < min_trial_id:
            return False
        if max_trial_id is not None:
          if trial.id > max_trial_id:
            return False
        if status_matches:
          if trial.status != status_matches:
            return False
      return True

    filtered_pytrials = [t for t in all_pytrials if _trial_filter(t)]

    # Doesn't affect datastore when measurements are deleted.
    if not include_intermediate_measurements:
      for filtered_pytrial in filtered_pytrials:
        filtered_pytrial.measurements = []

    return filtered_pytrials

  def CheckCancelled(self, note: Optional[str] = None) -> None:
    """Throws a CancelComputeError on timeout or if Vizier cancels."""
    pass  # Do nothing since it's one single process.

  def TimeRemaining(self) -> datetime.timedelta:
    """The time remaining to compute a result."""
    return datetime.timedelta.max  # RPCs don't have timeouts in OSS.

  def SendMetadata(self, delta: pythia.MetadataDelta) -> None:
    """Updates the metadata."""
    self.CheckCancelled('UpdateMetadata entry')
    request = vizier_service_pb2.UpdateMetadataRequest(name=self._study_guid)
    # Study Metadata
    for ns in delta.on_study.namespaces():
      for k, v in delta.on_study.abs_ns(ns).items():
        utils.AssignKeyValuePlus(request, trial_id=None, key=k, ns=ns, value=v)
    # Trial Metadata
    for trial_id, trial_metadata in delta.on_trials.items():
      for ns in trial_metadata.namespaces():
        for k, v in trial_metadata.abs_ns(ns).items():
          utils.AssignKeyValuePlus(
              request, trial_id=trial_id, key=k, ns=ns, value=v)
    waiter = utils.ResponseWaiter()
    # In a real server, this happens in another thread:
    response = self._vizier_service.UpdateMetadata(request, None)
    waiter.Report(response)
    # Wait for Vizier to reply with a UpdateMetadataResponse packet.
    waiter.WaitForResponse()
    self.CheckCancelled('UpdateMetadata exit')



"""Base class for all PythiaPolicies."""

import abc
from collections import abc as cabc
from typing import Any, FrozenSet, Iterable, List, Optional, Type, TypeVar

import attr
from vizier import pyvizier as vz

_T = TypeVar('_T')


def _is_positive(instance: Any, attribute: Any, value: Any):
  del instance, attribute
  if value <= 0:
    raise ValueError(f'value must be positive! given: {value}')


def _not_empty(instance: Any, attribute, value: str):
  del instance, attribute
  if not value:
    raise ValueError(f'value must be not nullable! given: {value}')


@attr.define
class EarlyStopDecision:
  """Stopping decision on a single trial.

  Attributes:
    id: Trial's id.
    reason: Explanations for how `should_stop` value was determined.
    should_stop:
    metadata:
    predicted_final_measurement: This is added to the Trial so that other
      algorithms can treat it as if it is the observed final measurement.
  """
  id: int = attr.ib(
      validator=[attr.validators.instance_of(int), _is_positive],
      on_setattr=attr.setters.validate)
  # TODO: Record this in DB even when `should_stop` == False.
  reason: str = attr.ib(
      validator=[attr.validators.instance_of(str), _not_empty],
      on_setattr=attr.setters.validate,
      converter=str)
  should_stop: bool = attr.ib(
      default=True,
      validator=[attr.validators.instance_of(bool)],
      on_setattr=attr.setters.validate)
  metadata: vz.Metadata = attr.ib(
      default=attr.Factory(vz.Metadata),
      validator=[attr.validators.instance_of(vz.Metadata)],
      on_setattr=attr.setters.validate)

  # TODO: Add a proper support for this in the service side.
  predicted_final_measurement: Optional[vz.Measurement] = attr.ib(
      default=None,
      validator=attr.validators.optional(
          attr.validators.instance_of(vz.Measurement)),
      on_setattr=attr.setters.validate)


@attr.define
class EarlyStopRequest:
  """Early stopping request.

  Attributes:
    study_guid:
    trial_ids: Trials to be considered for stopping. Used as hints.
    study_config:
    checkpoint_dir: If the policy wishes to use a checkpoint, then this is the
      path to find one.
  """
  _study_descriptor: vz.StudyDescriptor = attr.field(
      validator=attr.validators.instance_of(vz.StudyDescriptor))
  trial_ids: FrozenSet[int] = attr.field(
      default=attr.Factory(frozenset),
      validator=attr.validators.instance_of(FrozenSet),
      converter=frozenset)

  checkpoint_dir: Optional[str] = attr.field(
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)))

  @property
  def study_guid(self) -> str:
    return self._study_descriptor.guid

  @property
  def study_config(self) -> vz.StudyConfig:
    return self._study_descriptor.config


@attr.define(init=True)
class SuggestDecision:
  """Suggest decision.

  Currently only supports suggesting new points to be evaluated. In the future,
  we want to support other operations such as: freeze/thaw, or warm-start.

  Attributes:
    parameters: Trial parameters to be newly evaluated.
    metadata:
  """
  parameters: vz.ParameterDict = attr.field(
      init=True,
      validator=attr.validators.instance_of(vz.ParameterDict),
      converter=vz.ParameterDict)
  metadata: vz.Metadata = attr.field(
      init=True,
      factory=vz.Metadata,
      validator=[attr.validators.instance_of(vz.Metadata)])

  def to_trial(self, trial_id: int) -> vz.Trial:
    return vz.Trial(
        id=trial_id, parameters=self.parameters, metadata=self.metadata)


class SuggestDecisions(cabc.Sequence[SuggestDecision]):
  """Sequence of suggestions."""

  def __init__(self, items: Iterable[SuggestDecision] = tuple()):
    self._container = tuple(items)

  def __len__(self) -> int:
    return len(self._container)

  def __getitem__(self, index: int) -> SuggestDecision:
    return self._container[index]

  # TODO: rename to create_new_trials.
  @classmethod
  def from_trials(cls: Type[_T], trials: Iterable[vz.TrialSuggestion]) -> _T:
    return [SuggestDecision(t.parameters, t.metadata) for t in trials]


@attr.define
class SuggestRequest:
  """Suggestion Request.

  Attributes:
    study_descriptor: information about the Study.
    study_guid: Study id
    count: A recommendation for how many suggestions should be generated.
    study_config:
    checkpoint_dir: (If set) A system-provided directory where the policy can
      store a checkpoint.
  """
  study_descriptor: vz.StudyDescriptor = attr.field(
      validator=attr.validators.instance_of(vz.StudyDescriptor),
      on_setattr=attr.setters.frozen)

  count: int = attr.field(
      validator=[attr.validators.instance_of(int), _is_positive],
      on_setattr=attr.setters.validate)

  checkpoint_dir: Optional[str] = attr.field(
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
      on_setattr=attr.setters.validate)

  @property
  def study_config(self) -> vz.StudyConfig:
    return self.study_descriptor.config

  def study_guid(self) -> str:
    return f'{self.study_descriptor.guid}'


class Policy(abc.ABC):
  """Interface for Pythia Policy subclasses.

  Most Policy subclasses would wish to take `PolicySupporter` object in the
  `__init__` function. `PolicySupporter` provides an abstraction for how
  `Policy` reads more information about the Study in question beyond the basic
  information available in `SuggestRequest` and `EarlyStopRequest`. It allows
  the `Policy` to be compatible in multiple environments.
  """

  @abc.abstractmethod
  def suggest(self, request: SuggestRequest) -> SuggestDecisions:
    """Compute suggestions that Vizier will eventually hand to the user.

    Args:
      request:

    Returns:
      A list of Trials that will be passed on to the user.
      (See caveats in the SuggestionAnswer proto.)

    Raises:
      TemporaryPythiaError:  Generic retryable error.
      InactivateStudyError:  Raise this to inactivate the Study (non-retryable
        error).
        E.g. if this Policy cannot handle this StudyConfig.
        E.g. if this StudyConfig is somehow invalid.
        E.g. if this no more suggestions will ever be generated.
      CachedPolicyIsStaleError: Causes the computation to be restarted with a
        freshly constructed Policy instance.  It is incorrect to raise
        this on the first use of a Policy; the Study will be inactivated.

    NOTE: Trials should not have Trial.id set; these are new Trials rather than
      modifications of existing Trials.
    """
    pass

  def early_stop(self, request: EarlyStopRequest) -> List[EarlyStopDecision]:
    """Decide which Trials Vizier should stop.

    This returns a list of decisions on on-going Trials.
    Args:
      request:

    Returns:
      List of length up to `len(request.trial_ids)`. No decision is treated
      as "do not stop".

    Raises:
      TemporaryPythiaError:  Generic retryable error.
      InactivateStudyError: If this Pythia is inappropriate for the StudyConfig.
        (Non-retryable error.)  E.g. raise this if your Policy does not
        support MakeEarlyEarlyStopDecisions().
      CachedPolicyIsStaleError: Causes the computation to be restarted with a
        freshly constructed Policy instance.  It is incorrect to raise
        this on the first use of a Policy; the Study will be inactivated.
    """
    del request
    return []

  @property
  def should_be_cached(self) -> bool:
    """Returns True if it's safe & worthwhile to cache this Policy in RAM.

    This is called after MakeEarlyEarlyStopDecisions() and/or MakeSuggestions().
    If True, the policy may be stored in RAM (at least for a while), and state
    may be preserved for the next time that Study makes it to that Pythia
    server.
    """
    return False



"""RPC functions implemented from vizier_service.proto."""

import collections
import datetime
import threading
from typing import Optional

from absl import logging
import grpc
import numpy as np
from vizier import pythia
from vizier import pyvizier as base_pyvizier
from vizier._src.algorithms.designers import emukit
from vizier._src.algorithms.evolution import nsga2
from vizier._src.algorithms.policies import designer_policy as dp
from vizier._src.algorithms.policies import random_policy
from vizier.service import datastore
from vizier.service import pyvizier
from vizier.service import resources
from vizier.service import service_policy_supporter
from vizier.service import study_pb2
from vizier.service import vizier_oss_pb2
from vizier.service import vizier_service_pb2
from vizier.service import vizier_service_pb2_grpc

from google.longrunning import operations_pb2
from google.protobuf import empty_pb2
from google.protobuf import timestamp_pb2
from google.rpc import code_pb2
from google.rpc import status_pb2


def policy_creator(
    algorithm: study_pb2.StudySpec.Algorithm,
    policy_supporter: service_policy_supporter.ServicePolicySupporter
) -> pythia.Policy:
  """Creates a policy."""
  if algorithm in (study_pb2.StudySpec.Algorithm.ALGORITHM_UNSPECIFIED,
                   study_pb2.StudySpec.Algorithm.RANDOM_SEARCH):
    return random_policy.RandomPolicy(policy_supporter)
  elif algorithm == study_pb2.StudySpec.Algorithm.NSGA2:
    return dp.PartiallySerializableDesignerPolicy(policy_supporter,
                                                  nsga2.create_nsga2)
  elif algorithm == study_pb2.StudySpec.Algorithm.EMUKIT_GP_EI:
    return dp.DesignerPolicy(policy_supporter, emukit.EmukitDesigner)
  else:
    raise ValueError(
        f'Algorithm {study_pb2.StudySpec.Algorithm.Name(algorithm)} '
        'is not registered.')


def _get_current_time() -> timestamp_pb2.Timestamp:
  now = timestamp_pb2.Timestamp()
  now.GetCurrentTime()
  return now


MAX_STUDY_ID = 2147483647  # Max int32 value.


# TODO: remove context = None
# TODO: remove context = None
class VizierService(vizier_service_pb2_grpc.VizierServiceServicer):
  """Implements the GRPC functions outlined in vizier_service.proto."""

  def __init__(
      self,
      early_stop_recycle_period: datetime.timedelta = datetime.timedelta(
          seconds=60)):
    """Initializes the service.

    Creates the datastore and relevant locks for multhreading. Note that the
    datastore input/output is assumed to always be pass-by-value.

    Args:
      early_stop_recycle_period: Amount of time needed to pass before recycling
        an early stopping operation. See `CheckEarlyStoppingState` for more
        details.
    """
    self.datastore = datastore.NestedDictRAMDataStore()

    # For database edits using owner names.
    self._owner_name_to_lock = collections.defaultdict(threading.Lock)
    # For database edits using study names.
    self._study_name_to_lock = collections.defaultdict(threading.Lock)

    self._early_stop_recycle_period = early_stop_recycle_period

  def CreateStudy(
      self,
      request: vizier_service_pb2.CreateStudyRequest,
      context: Optional[grpc.ServicerContext] = None) -> study_pb2.Study:
    """Creates a study or loads an existing one.

    The initial request contains the study without the study name, only
    a user/client specified display_name for locating a potential pre-existing
    study. The returned study will then have the service-provided resource name.

    Args:
      request: A CreateStudyRequest.
      context: Optional GRPC ServicerContext.

    Returns:
      study: The Study created with generated service resource name.
    """
    study = request.study
    owner_id = resources.OwnerResource.from_name(request.parent).owner_id
    if request.study.name:
      raise ValueError(
          'Study should not have a resource name. Study names can only be assigned by the Vizier service.'
      )
    if not request.study.display_name:
      raise ValueError('Study display_name must be specified.')

    with self._owner_name_to_lock[request.parent]:
      # Database creates a new active study or loads existing study using the
      # display name.
      try:
        possible_candidate_studies = self.datastore.list_studies(request.parent)
      except datastore.NotFoundError:
        possible_candidate_studies = []

      # Check if all possible study_id's have been taken.
      if len(possible_candidate_studies) >= MAX_STUDY_ID:
        raise ValueError(
            'Maximum number of studies reached for owner {}.'.format(owner_id))

      for candidate_study in possible_candidate_studies:
        if candidate_study.display_name == request.study.display_name:
          logging.info('Found existing study of owner=%s display_name=%s!',
                       owner_id, candidate_study.display_name)
          return candidate_study

      # No study in the database matches the resource name. Making a new one.
      # study_id must be unique among existing studies.
      study_id = study.display_name

      # Finally create study in database and return it.
      study.name = resources.StudyResource(owner_id, study_id).name
      self.datastore.create_study(study)
    return study

  def GetStudy(
      self,
      request: vizier_service_pb2.GetStudyRequest,
      context: Optional[grpc.ServicerContext] = None) -> study_pb2.Study:
    """Gets a Study by name. If the study does not exist, return error."""
    return self.datastore.load_study(request.name)

  def ListStudies(
      self,
      request: vizier_service_pb2.ListStudiesRequest,
      context: Optional[grpc.ServicerContext] = None
  ) -> vizier_service_pb2.ListStudiesResponse:
    """Lists all the studies in a region for an associated project."""
    list_of_studies = self.datastore.list_studies(request.parent)
    return vizier_service_pb2.ListStudiesResponse(studies=list_of_studies)

  def DeleteStudy(
      self,
      request: vizier_service_pb2.DeleteStudyRequest,
      context: Optional[grpc.ServicerContext] = None) -> empty_pb2.Empty:
    """Deletes a Study."""
    lock = self._owner_name_to_lock[resources.StudyResource.from_name(
        request.name).owner_resource.name]
    with lock:
      self.datastore.delete_study(request.name)
    return empty_pb2.Empty()

  def SuggestTrials(
      self,
      request: vizier_service_pb2.SuggestTrialsRequest,
      context: Optional[grpc.ServicerContext] = None
  ) -> operations_pb2.Operation:
    """Adds one or more Trials to a Study, with parameter values suggested by a Pythia policy.

    The logic is as follows:
    1. If there is already an active (not done) operation, simply return that.
    2. Else, create a new active operation.
    3. We need the requested number of ACTIVE trials to return. These will come
    from 3 sources:
      A. ACTIVE trials already assigned to the client.
      B. REQUESTED trials which not been assigned to any client.
      C. Pythia-computed suggestions.

    We first look from source A, then source B, then source C. If we ever reach
    source C and Pythia over-delivers too many suggestions (e.g. evolutionary
    algorithms sometimes do so because of batched behaviors), we put the extra
    suggestions into the REQUESTED pool (i.e. source B) for future use.

    The returned operation can either be:
    1. Done (Successfully contains the requested number of suggestions)
    2. Error-ed, which will happen if Pythia had an issue producing the right
    number of trials (via numerical crash or under-delivering).

    Args:
      request:
      context:

    Returns:
      A long-running operation associated with the generation of Trial
      suggestions. When this long-running operation succeeds, it will contain a
      [SuggestTrialsResponse].
    """
    # Convenient names and id's to be used below.
    study_name = request.parent
    study_resource = resources.StudyResource.from_name(study_name)
    study_id = study_resource.study_id
    owner_id = study_resource.owner_id
    owner_name = study_resource.owner_resource.name

    # Lock the database for this study and load it.
    with self._study_name_to_lock[request.parent]:
      study = self.datastore.load_study(request.parent)

      # Checks for a non-done operation in the database with this name.
      active_op_filter_fn = lambda op: not op.done
      try:
        active_op_list = self.datastore.list_suggestion_operations(
            owner_name, request.client_id, active_op_filter_fn)
      except datastore.NotFoundError:
        active_op_list = []
      if active_op_list:
        return active_op_list[0]  # We've found the active one!

      start_time = _get_current_time()
      # Create a new Op if there aren't any active (not done) ops.
      try:
        new_op_number = self.datastore.max_suggestion_operation_number(
            owner_name, request.client_id) + 1
      except datastore.NotFoundError:
        new_op_number = 1
      new_op_name = resources.SuggestionOperationResource(
          owner_id, request.client_id, new_op_number).name
      output_op = operations_pb2.Operation(name=new_op_name, done=False)
      self.datastore.create_suggestion_operation(output_op)

      # Check how many ACTIVE trials already exist for this client only.
      all_trials = self.datastore.list_trials(study_name)
      active_trials = [
          t for t in all_trials if t.state == study_pb2.Trial.State.ACTIVE and
          t.client_id == request.client_id
      ]
      if len(active_trials) >= request.suggestion_count:
        output_op.response.value = vizier_service_pb2.SuggestTrialsResponse(
            trials=active_trials[:request.suggestion_count],
            start_time=start_time).SerializeToString()
        output_op.done = True
        self.datastore.update_suggestion_operation(output_op)
        return output_op
      else:
        output_trials = active_trials

        # Get suggestions from the pool of requested trials.
        requested_trials = [
            t for t in all_trials if t.state == study_pb2.Trial.State.REQUESTED
        ]
        while request.suggestion_count - len(
            output_trials) > 0 and requested_trials:
          assigned_trial = requested_trials.pop()
          assigned_trial.state = study_pb2.Trial.State.ACTIVE
          assigned_trial.client_id = request.client_id
          assigned_trial.start_time.CopyFrom(start_time)
          self.datastore.update_trial(assigned_trial)
          output_trials.append(assigned_trial)

        if len(output_trials) == request.suggestion_count:
          # We've finished collecting enough trials from the REQUESTED pool.
          output_op.response.value = vizier_service_pb2.SuggestTrialsResponse(
              trials=output_trials, start_time=start_time).SerializeToString()
          output_op.done = True
          self.datastore.update_suggestion_operation(output_op)
          return output_op

        # Still need more suggestions. Pythia begins computing missing amount.
        policy_supporter = service_policy_supporter.ServicePolicySupporter(
            study_name, self)
        pythia_policy = policy_creator(study.study_spec.algorithm,
                                       policy_supporter)

        pythia_sc = pyvizier.StudyConfig.from_proto(study.study_spec)
        study_descriptor = base_pyvizier.StudyDescriptor(
            config=pythia_sc,
            guid=study_name,
            max_trial_id=self.datastore.max_trial_id(study_name))
        suggest_request = pythia.SuggestRequest(
            study_descriptor=study_descriptor,
            count=request.suggestion_count - len(output_trials))
        try:
          suggest_decisions = pythia_policy.suggest(suggest_request)
          assert len(suggest_decisions
                    ) >= request.suggestion_count - len(output_trials)
        # Leaving a broad catch for now since Pythia can raise any exception.
        except Exception as e:  # pylint: disable=broad-except
          output_op.error.CopyFrom(
              status_pb2.Status(code=code_pb2.Code.INTERNAL, message=str(e)))
          logging.exception(
              'Failed to request trials from Pythia for request: %s', request)
          output_op.done = True
          self.datastore.update_suggestion_operation(output_op)
          return output_op

        new_py_trials = [
            pyvizier.Trial(parameters=decision.parameters)
            for decision in suggest_decisions
        ]
        new_trials = pyvizier.TrialConverter.to_protos(new_py_trials)

        while request.suggestion_count - len(output_trials) > 0:
          new_trial = new_trials.pop()
          trial_id = self.datastore.max_trial_id(request.parent) + 1
          new_trial.id = str(trial_id)
          new_trial.name = resources.TrialResource(owner_id, study_id,
                                                   trial_id).name
          new_trial.state = study_pb2.Trial.State.ACTIVE
          new_trial.start_time.CopyFrom(start_time)
          new_trial.client_id = request.client_id
          self.datastore.create_trial(new_trial)
          output_trials.append(new_trial)

        output_op.response.value = vizier_service_pb2.SuggestTrialsResponse(
            trials=output_trials, start_time=start_time).SerializeToString()

        # Store remaining trials as REQUESTED if Pythia over-delivered.
        for remaining_trial in new_trials:
          trial_id = self.datastore.max_trial_id(request.parent) + 1
          remaining_trial.id = str(trial_id)
          remaining_trial.name = resources.TrialResource(
              owner_id, study_id, trial_id).name
          remaining_trial.state = study_pb2.Trial.State.REQUESTED
          self.datastore.create_trial(new_trial)

        output_op.done = True
        self.datastore.update_suggestion_operation(output_op)
        return output_op

  def GetOperation(
      self,
      request: operations_pb2.GetOperationRequest,
      context: Optional[grpc.ServicerContext] = None
  ) -> operations_pb2.Operation:
    """Gets the latest state of a SuggestTrials() long-running operation."""
    operation = self.datastore.get_suggestion_operation(request.name)
    return operation

  def CreateTrial(
      self,
      request: vizier_service_pb2.CreateTrialRequest,
      context: Optional[grpc.ServicerContext] = None) -> study_pb2.Trial:
    """Adds user provided Trial to a Study and assigns the correct fields."""
    with self._study_name_to_lock[request.parent]:
      trial = request.trial
      trial.id = str(self.datastore.max_trial_id(request.parent) + 1)
      trial.name = (resources.StudyResource.from_name(
          request.parent).trial_resource(trial_id=trial.id)).name

      if trial.state != study_pb2.Trial.State.SUCCEEDED:
        trial.state = study_pb2.Trial.State.REQUESTED
      trial.ClearField('client_id')

      trial.start_time.CopyFrom(_get_current_time())
      self.datastore.create_trial(trial)
      return trial

  def GetTrial(self, request: vizier_service_pb2.GetTrialRequest,
               context: grpc.ServicerContext) -> Optional[study_pb2.Trial]:
    """Gets a Trial."""
    try:
      return self.datastore.get_trial(request.name)
    except datastore.NotFoundError as e:
      context.set_code(grpc.StatusCode.NOT_FOUND)
      context.set_details(str(e))

  def ListTrials(
      self,
      request: vizier_service_pb2.ListTrialsRequest,
      context: Optional[grpc.ServicerContext] = None
  ) -> vizier_service_pb2.ListTrialsResponse:
    """Lists the Trials associated with a Study."""
    list_of_trials = self.datastore.list_trials(request.parent)
    return vizier_service_pb2.ListTrialsResponse(trials=list_of_trials)

  def AddTrialMeasurement(
      self,
      request: vizier_service_pb2.AddTrialMeasurementRequest,
      context: Optional[grpc.ServicerContext] = None) -> study_pb2.Trial:
    """Adds a measurement of the objective metrics to a Trial.

    This measurement is assumed to have been taken before the Trial is
    complete.

    Args:
      request:
      context:

    Returns:
      Trial whose measurement was appended.
    """
    trial = self.datastore.get_trial(request.trial_name)
    trial.measurements.extend([request.measurement])
    self.datastore.update_trial(trial)
    return trial

  # TODO: Auto selection defaults to the last measurement.
  # Add support for "best measurement" behavior.
  def CompleteTrial(
      self,
      request: vizier_service_pb2.CompleteTrialRequest,
      context: Optional[grpc.ServicerContext] = None) -> study_pb2.Trial:
    """Marks a Trial as complete."""
    trial = self.datastore.get_trial(request.name)

    if request.final_measurement.metrics:
      trial.final_measurement.CopyFrom(request.final_measurement)
      trial.state = study_pb2.Trial.State.SUCCEEDED
    elif not request.trial_infeasible:
      # Trial's final measurement auto-selected from latest reported
      # measurement.
      trial.state = study_pb2.Trial.State.SUCCEEDED
      if trial.measurements:
        trial.final_measurement.CopyFrom(trial.measurements[-1])
      else:
        raise ValueError(
            "Both the request and trial intermediate measurements are missing. Cannot determine trial's final_measurement."
        )

    # Handle infeasibility.
    if request.trial_infeasible:
      trial.state = study_pb2.Trial.State.INFEASIBLE
      trial.infeasible_reason = request.infeasible_reason

    self.datastore.update_trial(trial)
    return trial

  def DeleteTrial(
      self,
      request: vizier_service_pb2.DeleteTrialRequest,
      context: Optional[grpc.ServicerContext] = None) -> empty_pb2.Empty:
    """Deletes a Trial."""
    self.datastore.delete_trial(request.name)
    return empty_pb2.Empty()

  # TODO: This curerntly uses the same algorithm as suggestion.
  def CheckTrialEarlyStoppingState(
      self,
      request: vizier_service_pb2.CheckTrialEarlyStoppingStateRequest,
      context: Optional[grpc.ServicerContext] = None
  ) -> vizier_service_pb2.CheckTrialEarlyStoppingStateResponse:
    """Checks whether a Trial should stop or not.

    The logic is as follows:

    1. We check if there already exists an operation. If not, create it. But if
    it already exists:
      a. If it's still ACTIVE (i.e. early stopping is still being computed by
      Pythia), just return it.
      b. If it's done and the call is very recent, just return it.
      c. If it's been a long time, we can recycle the operation, and remark it
      as ACTIVE to be used again.

    2. Send the operation to Pythia, which will return a set of early stopping
    decisions. The Pythia policy is guaranteed to return a decision for the
    requested trial. However, Pythia policy may also signal whether other trials
    should stop or not (e.g. in algorithms in which proposed trials are batched
    and correlated).

    3. Transfer these decisions' `should_stop` into the relevant operations
    (some of which will also be created in this call, for future use), including
    the original operation which needs to be returned. Mark them all as done.

    4. Return the original trial's `should_stop` decision.

    Args:
      request:
      context:

    Returns:
      A CheckTrialEarlyStoppingStateResponse, containing a bool to denote
      whether the requested trial should stop.
    """
    trial_resource = resources.TrialResource.from_name(request.trial_name)
    study_name = trial_resource.study_resource.name
    outer_op_name = resources.EarlyStoppingOperationResource(
        trial_resource.owner_id, trial_resource.study_id,
        trial_resource.trial_id).name

    with self._study_name_to_lock[study_name]:
      try:
        # Reuse any existing early stopping op, since the Pythia policy may have
        # already signaled this trial to stop.
        output_operation = self.datastore.get_early_stopping_operation(
            outer_op_name)
      except KeyError:
        output_operation = None

      if output_operation is None:
        # Create fresh new operation.
        output_operation = vizier_oss_pb2.EarlyStoppingOperation(
            name=outer_op_name,
            status=vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE,
            should_stop=False)
        output_operation.creation_time.CopyFrom(_get_current_time())
        self.datastore.create_early_stopping_operation(output_operation)

      else:
        if output_operation.status == vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE:
          # Operation is already active. Just return it.
          return vizier_service_pb2.CheckTrialEarlyStoppingStateResponse(
              should_stop=output_operation.should_stop)
        elif datetime.datetime.utcnow(
        ) - output_operation.completion_time.ToDatetime(
        ) < self._early_stop_recycle_period:
          # Operation is very recent. Just return it.
          return vizier_service_pb2.CheckTrialEarlyStoppingStateResponse(
              should_stop=output_operation.should_stop)
        else:
          # Recycle the operation to ACTIVE again and start Pythia for
          # recomputation.
          output_operation.status = vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE
          output_operation.should_stop = False  # Defaulted back to False.
          self.datastore.update_early_stopping_operation(output_operation)

      study = self.datastore.load_study(study_name)
      policy_supporter = service_policy_supporter.ServicePolicySupporter(
          study_name, self)
      pythia_policy = policy_creator(study.study_spec.algorithm,
                                     policy_supporter)
      pythia_sc = pyvizier.StudyConfig.from_proto(study.study_spec)
      study_descriptor = base_pyvizier.StudyDescriptor(
          config=pythia_sc,
          guid=study_name,
          max_trial_id=self.datastore.max_trial_id(study_name))
      early_stop_request = pythia.EarlyStopRequest(
          study_descriptor=study_descriptor,
          trial_ids=[trial_resource.trial_id])
      early_stopping_decisions = pythia_policy.early_stop(early_stop_request)

      # Pythia does not guarantee that the output_operation's id
      # will be in the decisions.
      for early_stopping_decision in early_stopping_decisions:
        inner_op_name = resources.EarlyStoppingOperationResource(
            trial_resource.owner_id, trial_resource.study_id,
            early_stopping_decision.id).name
        try:
          inner_operation = self.datastore.get_early_stopping_operation(
              inner_op_name)
        except KeyError:
          # Create the operation to store early stopping data for future use.
          inner_operation = vizier_oss_pb2.EarlyStoppingOperation(
              name=inner_op_name,
              status=vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE,
              should_stop=False)
          inner_operation.creation_time.CopyFrom(_get_current_time())
          self.datastore.create_early_stopping_operation(inner_operation)

        inner_operation.should_stop = early_stopping_decision.should_stop
        inner_operation.status = vizier_oss_pb2.EarlyStoppingOperation.Status.DONE
        inner_operation.completion_time.CopyFrom(_get_current_time())
        self.datastore.update_early_stopping_operation(inner_operation)

      # Operation to be outputted may have changed.
      output_operation = self.datastore.get_early_stopping_operation(
          output_operation.name)
      return vizier_service_pb2.CheckTrialEarlyStoppingStateResponse(
          should_stop=output_operation.should_stop)

  def StopTrial(
      self,
      request: vizier_service_pb2.StopTrialRequest,
      context: Optional[grpc.ServicerContext] = None) -> study_pb2.Trial:
    trial = self.datastore.get_trial(request.name)
    trial.state = study_pb2.Trial.STOPPING
    self.datastore.update_trial(trial)
    return trial

  def ListOptimalTrials(
      self,
      request: vizier_service_pb2.ListOptimalTrialsRequest,
      context: Optional[grpc.ServicerContext] = None
  ) -> vizier_service_pb2.ListOptimalTrialsResponse:
    """The definition of pareto-optimal can be checked in wiki page.

    https://en.wikipedia.org/wiki/Pareto_efficiency.

    Args:
      request:
      context:

    Returns:
      A list containing pareto-optimal Trials for multi-objective Study or the
      optimal Trials for single-objective Study.
    """
    raw_trial_list = self.datastore.list_trials(request.parent)
    if not raw_trial_list:
      return vizier_service_pb2.ListOptimalTrialsResponse(optimal_trials=[])

    study_spec = self.datastore.load_study(request.parent).study_spec

    metric_id_to_goal = {m.metric_id: m.goal for m in study_spec.metrics}
    required_metric_ids = set(metric_id_to_goal.keys())

    considered_trials = []
    considered_trial_objective_vectors = []

    for trial in raw_trial_list:
      trial_metric_id_to_value = {
          m.metric_id: m.value for m in trial.final_measurement.metrics
      }
      trial_metric_ids = set(trial_metric_id_to_value.keys())
      # Add trials ONLY if they succeeded and contain all supposed metrics.
      if trial.state == study_pb2.Trial.State.SUCCEEDED and required_metric_ids.issubset(
          trial_metric_ids):
        objective_vector = []
        for metric_id, goal in metric_id_to_goal.items():
          # Flip sign for convenience when computing optimality.
          if goal == study_pb2.StudySpec.MetricSpec.GoalType.MINIMIZE:
            vector_value = -1.0 * trial_metric_id_to_value[metric_id]
          else:
            vector_value = trial_metric_id_to_value[metric_id]
          objective_vector.append(vector_value)

        considered_trials.append(trial)
        considered_trial_objective_vectors.append(objective_vector)

    if not considered_trials:
      return vizier_service_pb2.ListOptimalTrialsResponse(optimal_trials=[])

    # Find Pareto optimal trials.
    ys = np.array(considered_trial_objective_vectors)
    n = ys.shape[0]
    dominated = np.asarray(
        [[np.all(ys[i] <= ys[j]) & np.any(ys[j] > ys[i])
          for i in range(n)]
         for j in range(n)])
    optimal_booleans = np.logical_not(np.any(dominated, axis=0))
    optimal_trials = []
    for i, boolean in enumerate(list(optimal_booleans)):
      if boolean:
        optimal_trials.append(considered_trials[i])

    return vizier_service_pb2.ListOptimalTrialsResponse(
        optimal_trials=optimal_trials)

  def UpdateMetadata(
      self,
      request: vizier_service_pb2.UpdateMetadataRequest,
      context: Optional[grpc.ServicerContext] = None
  ) -> vizier_service_pb2.UpdateMetadataResponse:
    """Stores the supplied metadata in the database."""
    # TODO: Add locking logic.
    try:
      self.datastore.update_metadata(
          request.name,
          [x.k_v for x in request.metadata if not x.HasField('trial_id')],
          [x for x in request.metadata if x.HasField('trial_id')])
    except KeyError as e:
      return vizier_service_pb2.UpdateMetadataResponse(
          error_details=';'.join(e.args))
    return vizier_service_pb2.UpdateMetadataResponse()



"""Tests for vizier.service.vizier_client."""

# TODO: Rewrite using RPC calls and remove all direct lookups
# into the datastore.
# TODO: Cover delete_trial, delete_study, get_study_config, and
# add_trial, OR turn it into a private module

from concurrent import futures
import datetime
import time
from typing import List
from absl import logging
import grpc
import portpicker

from vizier.service import pyvizier
from vizier.service import resources
from vizier.service import study_pb2
from vizier.service import vizier_client
from vizier.service import vizier_server
from vizier.service import vizier_service_pb2_grpc

from absl.testing import absltest
from absl.testing import parameterized


class VizierClientTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    # Setup Vizier Service and some pre-stored data.
    self.early_stop_recycle_period = datetime.timedelta(seconds=1)
    self.servicer = vizier_server.VizierService(
        early_stop_recycle_period=self.early_stop_recycle_period)
    self.owner_id = 'my_username'
    self.study_id = '1231232'
    self.study_resource_name = resources.StudyResource(self.owner_id,
                                                       self.study_id).name
    self.client_id = 'my_client'

    double_value_spec = study_pb2.StudySpec.ParameterSpec.DoubleValueSpec(
        min_value=-1.0, max_value=1.0)
    double_parameter_spec = study_pb2.StudySpec.ParameterSpec(
        parameter_id='double', double_value_spec=double_value_spec)

    metric_spec = study_pb2.StudySpec.MetricSpec(
        metric_id='example_metric',
        goal=study_pb2.StudySpec.MetricSpec.GoalType.MAXIMIZE)
    self.example_study = study_pb2.Study(
        name=self.study_resource_name,
        study_spec=study_pb2.StudySpec(
            algorithm=study_pb2.StudySpec.Algorithm.RANDOM_SEARCH,
            parameters=[double_parameter_spec],
            metrics=[metric_spec]))

    self.active_trial = study_pb2.Trial(
        name=resources.TrialResource(self.owner_id, self.study_id, 1).name,
        id='1',
        state=study_pb2.Trial.State.ACTIVE)

    self.servicer.datastore.create_study(self.example_study)
    self.servicer.datastore.create_trial(self.active_trial)

    # Setup local networking.
    self.port = portpicker.pick_unused_port()
    self.address = f'localhost:{self.port}'

    # Setup server.
    self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=100))

    vizier_service_pb2_grpc.add_VizierServiceServicer_to_server(
        self.servicer, self.server)
    self.server.add_secure_port(self.address, grpc.local_server_credentials())
    self.server.start()

    # Setup connection to server.
    self.client = vizier_client.VizierClient.from_endpoint(
        service_endpoint=self.address,
        study_resource_name=self.study_resource_name,
        client_id=self.client_id)

  def test_create_or_load_study(self):
    study_config = pyvizier.StudyConfig()
    study_id = 'example_display_name'

    client = vizier_client.create_or_load_study(
        service_endpoint=self.address,
        owner_id=self.owner_id,
        client_id='a_client',
        study_id=study_id,
        study_config=study_config)
    study = self.servicer.datastore.load_study(client.study_resource_name)
    self.assertEqual(study.study_spec, study_config.to_proto())
    self.assertIsNotNone(study.name)

    another_client = vizier_client.create_or_load_study(
        service_endpoint=self.address,
        owner_id=self.owner_id,
        client_id='another_client',
        study_id=study_id,
        study_config=study_config)
    self.assertEqual(client.study_resource_name,
                     another_client.study_resource_name)

  def test_list_studies(self):
    study_list_json = self.client.list_studies()
    self.assertLen(study_list_json, 1)

  def test_delete_study(self):
    self.client.delete_study()
    empty_list_json = self.client.list_studies()
    self.assertEmpty(empty_list_json)

  def test_list_trials(self):
    trial_list = self.client.list_trials()
    self.assertLen(trial_list, 1)

  def test_list_optimal_trials(self):
    for i in range(2, 10):
      completed_trial = study_pb2.Trial(
          name=resources.TrialResource(self.owner_id, self.study_id, i).name,
          id=str(i),
          state=study_pb2.Trial.State.SUCCEEDED,
          final_measurement=study_pb2.Measurement(metrics=[
              study_pb2.Measurement.Metric(
                  metric_id='example_metric', value=0.2 * i)
          ]))

      self.servicer.datastore.create_trial(completed_trial)

    trial_list = self.client.list_optimal_trials()
    self.assertLen(trial_list, 1)

  def test_get_trial(self):
    active_trial = self.client.get_trial(trial_id=1)
    self.assertEqual(active_trial,
                     pyvizier.TrialConverter.from_proto(self.active_trial))

  @parameterized.named_parameters(('Infeasible', 'infeasible_reason'),
                                  ('Complete', None))
  def test_complete_trial(self, infeasibility_reason):
    final_measurement = pyvizier.Measurement(
        metrics={'metric': pyvizier.Metric(value=0.1)})
    output_trial = self.client.complete_trial(
        trial_id=1,
        final_measurement=final_measurement,
        infeasibility_reason=infeasibility_reason)

    self.assertEqual(output_trial.status, pyvizier.TrialStatus.COMPLETED)
    self.assertEqual(output_trial.infeasibility_reason, infeasibility_reason)

    # See if the rest of the contents were maintained.
    completed_trial = pyvizier.TrialConverter.from_proto(
        self.servicer.datastore.get_trial(self.active_trial.name))
    self.assertEqual(output_trial, completed_trial)

  def test_should_trial_stop(self):
    # Only trial 1 was ACTIVE, so RandomPolicy will signal to stop it.
    should_stop = self.client.should_trial_stop(trial_id=1)
    self.assertTrue(should_stop)
    self.client.stop_trial(trial_id=1)

    # The op will become recycled after the time period and early stopping will
    # be recomputed again. But RandomPolicy will consider the trial non-ACTIVE
    # and simply return False.
    time.sleep(self.early_stop_recycle_period.total_seconds())
    should_stop_again = self.client.should_trial_stop(trial_id=1)
    self.assertFalse(should_stop_again)

  def test_intermediate_measurement(self):
    self.client.report_intermediate_objective_value(
        step=5,
        elapsed_secs=3.0,
        metric_list=[{
            'example_metric': 5
        }],
        trial_id=1)

  def test_get_suggestions(self):
    suggestion_count = 2
    suggestions_list = self.client.get_suggestions(
        suggestion_count=suggestion_count)
    self.assertLen(suggestions_list, suggestion_count)
    logging.info('Suggestions List: %s', suggestions_list)

  @parameterized.parameters((pyvizier.Algorithm.RANDOM_SEARCH, 50, 1, False),
                            (pyvizier.Algorithm.EMUKIT_GP_EI, 7, 2, False),
                            (pyvizier.Algorithm.NSGA2, 50, 1, True))
  def test_e2e_tuning(self, algorithm, num_iterations: int, batch_size: int,
                      multi_objective: bool):
    # Runs end-to-end tuning via back-and-forth communication to server.
    def learning_curve_generator(learning_rate: float) -> List[float]:
      return [learning_rate * step for step in range(10)]

    study_config = pyvizier.StudyConfig()
    study_config.search_space.select_root().add_float_param(
        'learning_rate', min_value=0.0, max_value=1.0, default_value=0.5)
    study_config.search_space.select_root().add_int_param(
        'num_layers', min_value=1, max_value=5)
    study_config.metric_information = [
        pyvizier.MetricInformation(
            name='accuracy', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)
    ]
    if multi_objective:
      study_config.metric_information.append(
          pyvizier.MetricInformation(
              name='latency', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    study_config.algorithm = algorithm

    cifar10_client = vizier_client.create_or_load_study(
        service_endpoint=self.address,
        owner_id=self.owner_id,
        study_id='cifar10',
        study_config=study_config,
        client_id=self.client_id)

    for _ in range(num_iterations):
      suggestions = cifar10_client.get_suggestions(suggestion_count=batch_size)
      for trial in suggestions:
        learning_rate = trial.parameters.get_value('learning_rate')
        num_layers = trial.parameters.get_value('num_layers')
        curve = learning_curve_generator(learning_rate)
        for i in range(len(curve)):
          cifar10_client.report_intermediate_objective_value(
              step=i,
              elapsed_secs=0.1 * i,
              metric_list=[{
                  'accuracy': curve[i],
                  'latency': 0.5 * num_layers
              }],
              trial_id=trial.id)
        cifar10_client.complete_trial(trial_id=trial.id)

        # Recover the trial from database.
        study_resource = resources.StudyResource.from_name(
            cifar10_client.study_resource_name)
        trial_name = resources.TrialResource(study_resource.owner_id,
                                             study_resource.study_id,
                                             trial.id).name
        stored_trial = self.servicer.datastore.get_trial(trial_name)
        stored_curve = [m.metrics[0].value for m in stored_trial.measurements]

        # See if curve was stored correctly.
        self.assertEqual(curve, stored_curve)

        # See if final_measurement is defaulted to end of curve.
        final_accuracy = stored_trial.final_measurement.metrics[0].value
        self.assertEqual(curve[-1], final_accuracy)


if __name__ == '__main__':
  absltest.main()



"""Tests for local_service."""

from vizier.service.testing import local_service
from absl.testing import absltest


class LocalServiceTest(absltest.TestCase):

  def test_creation(self):
    service = local_service.LocalVizierTestService()
    self.assertIsNotNone(service.stub)
    self.assertIsNotNone(service.endpoint)
    self.assertIsNotNone(service.datastore)


if __name__ == '__main__':
  absltest.main()



import numpy as np

from vizier._src.pyvizier.multimetric import hypervolume
from absl.testing import absltest


class HypervolumeTest(absltest.TestCase):

  def testParetoHypervolume(self):
    x = np.random.normal()
    y = np.random.normal()
    points = np.array([[x, y], [y, x]])
    origin = np.array([-1, -1])
    pf = hypervolume.ParetoFrontier(points, origin)
    self.assertAlmostEqual(
        pf.hypervolume(),
        max(2 * (x + 1) * (y + 1) - min(x + 1, y + 1)**2, 0.0),
        delta=0.1)

  def testParetoHypervolumeCumulative(self):
    points = np.random.normal(size=(100, 4))
    origin = np.zeros(4)

    pf = hypervolume.ParetoFrontier(points, origin)
    cumulative_vol = pf.hypervolume(is_cumulative=True)
    self.assertLen(cumulative_vol, len(points))
    for i in range(len(points)):
      if i < len(points) - 1:
        self.assertGreaterEqual(cumulative_vol[i + 1], cumulative_vol[i])
      else:
        self.assertAlmostEqual(cumulative_vol[i], pf.hypervolume())

  def testDimensionMismatchInit(self):
    points = np.array([[2, 1], [1, 2], [0, 0]])
    origin_3d = np.array([-1, -1, 1])
    # Dimension mismatch.
    with self.assertRaisesRegex(ValueError, 'Dimension mismatch'):
      hypervolume.ParetoFrontier(points, origin_3d)

  def testDimensionMismatchHypervolume(self):
    points = np.array([[2, 1], [1, 2], [0, 0]])
    origin_3d = np.array([-1, -1, 1])
    pf = hypervolume.ParetoFrontier(points, origin=np.array([0, 0]))
    # Dimension mismatch.
    with self.assertRaisesRegex(ValueError, 'Dimension mismatch'):
      pf.hypervolume(additional_points=origin_3d[..., np.newaxis])


if __name__ == '__main__':
  absltest.main()



"""Cross-platform Vizier client interfaces.

Aside from "materialize_" methods, code written using these interfaces are
compatible with OSS and Cloud Vertex Vizier. Note importantly that subclasses
may have more methods than what is required by interfaces, and such methods
are not cross compatible. Our recommendation is to explicitly type your objects
to be `StudyInterface` or `TrialInterface` when you want to guarantee that
a code block is cross-platform.

Keywords:

#Materialize: The method returns a deep copy of the underlying pyvizier object.
Modifying the returned object does not update the Vizier service.
"""

# TODO: Add a dedicated .md file with more code examples.

import abc

from typing import Iterable, Optional, Collection, Type, TypeVar, Mapping, Any

from vizier import pyvizier as vz

_T = TypeVar('_T')


class ResourceNotFoundError(LookupError):
  """Error raised by Vizier clients when resource is not found."""
  pass


class TrialInterface(abc.ABC):
  """Responsible for trial-level operations."""

  @property
  @abc.abstractmethod
  def id(self) -> int:
    """Identifier of the trial within the study.

    Ids are sorted in increasing order of creation time.
    """

  @property
  @abc.abstractmethod
  def parameters(self) -> Mapping[str, Any]:
    """#Materializes the parameters of the trial."""

  @abc.abstractmethod
  def delete(self) -> None:
    """Delete the Trial in Vizier service.

    There is currently no promise on how this object behaves after `delete()`.
    If you are sharing a Trial object in parallel processes, proceed with
    caution.
    """

  @abc.abstractmethod
  def complete(
      self,
      measurement: Optional[vz.Measurement] = None,
      *,
      infeasible_reason: Optional[str] = None) -> Optional[vz.Measurement]:
    """Completes the trial and #materializes the measurement.

    * If `measurement` is provided, then Vizier writes it as the trial's final
    measurement and returns it.
    * If `infeasible_reason` is provided, `measurement` is not required but
    can still be specified.
    * If neither is provided, then Vizier selects an existing (intermediate)
    measurement to be the final measurement and returns it.

    Args:
      measurement: Final measurement.
      infeasible_reason: Infeasible reason for missing final measurement.

    Returns:
      The final measurement of the trial, or None if the trial is marked
      infeasible.

    Raises:
      ValueError: If neither `measurement` nor `infeasible_reason` is provided
        but the trial does not contain any intermediate measurements.
    """

  @abc.abstractmethod
  def check_early_stopping(self) -> bool:
    """Decide if the trial should stop.

    If the Trial is in ACTIVE state and an early stopping algorithm is
    configured fior the Study, the stopping algorithm makes a stopping decision.
    If the algorithm decides that the trial is not worth continuing, Vizier
    service moves it into `STOPPING` state.

    Then, this method returns True if the Trial is in `STOPPING` state.

    Returns:
      True if trial is already in STOPPING state or entered STOPPING as a
      result of this method invocation.
    """
    pass

  @abc.abstractmethod
  def add_measurement(self, measurement: vz.Measurement) -> None:
    """Adds an intermediate measurement."""

  @abc.abstractmethod
  def materialize(self, *, include_all_measurements: bool = True) -> vz.Trial:
    """#Materializes the Trial.

    Args:
      include_all_measurements: If True, returned Trial includes all
        intermediate measurements. The `final_measurement` is always included if
        one exists.

    Returns:
      Trial object.
    """


class StudyInterface(abc.ABC):
  """Responsible for study-level operations."""

  @property
  @abc.abstractmethod
  def resource_name(self) -> str:
    """Globally unique identifier of the study in Vizier service."""

  @abc.abstractmethod
  def suggest(
      self,
      *,
      count: Optional[int] = None,
      client_id: str = 'default_client_id') -> Collection[TrialInterface]:
    """Returns Trials to be evaluated by client_id.

    Args:
      count: Number of suggestions.
      client_id: When new Trials are generated, their `client_id` field is
        populated with this client_id. The Vizier service first looks for
        existing ACTIVE Trials that are assigned to `client_id`, before
        generating new ones.

    Returns:
      Trials.
    """

  @abc.abstractmethod
  def delete(self) -> None:
    """Delete the Study in Vizier service.

    There is currently no promise on how this object behaves after `delete()`.
    If you are sharing a Study object in parallel processes, proceed with
    caution.
    """

  # TODO: Make this method public.
  @abc.abstractmethod
  def _add_trial(self, trial: vz.Trial) -> TrialInterface:
    """Adds a trial to the Study. For testing only."""

  @abc.abstractmethod
  def trials(
      self,
      trial_filter: Optional[vz.TrialFilter] = None
  ) -> Iterable[TrialInterface]:
    """Fetches a collection of trials."""

  @abc.abstractmethod
  def get_trial(self, uid: int) -> TrialInterface:
    """Fetches a single trial.

    Args:
      uid: Unique identifier of the trial within study.

    Returns:
      Trial.

    Raises:
      ResourceNotFoundError: If trial does not exist.
    """

  @abc.abstractmethod
  def optimal_trials(self) -> Iterable[TrialInterface]:
    """Returns optimal trial(s)."""

  @abc.abstractmethod
  def materialize_problem_statement(self) -> vz.ProblemStatement:
    """#Materializes the problem statement."""

  @classmethod
  @abc.abstractmethod
  def from_resource_name(cls: Type[_T], name: str, /) -> _T:
    """Fetches an existing study from the Vizier service.

    Args:
      name: Globally unique identifier of the study.

    Returns:
      Study.

    Raises:
      ResourceNotFoundError: If study does not exist.
    """
